# sage_setup: distribution = sagemath-combinat
"""
Recursive species
"""
# ****************************************************************************
#       Copyright (C) 2008 Mike Hansen <mhansen@gmail.com>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
# ****************************************************************************
from sage.combinat.species.species import GenericCombinatorialSpecies
from sage.combinat.species.structure import SpeciesStructureWrapper
from sage.rings.rational_field import QQ


class CombinatorialSpeciesStructure(SpeciesStructureWrapper):
    pass


class CombinatorialSpecies(GenericCombinatorialSpecies):
    def __init__(self, min=None):
        """
        EXAMPLES::

            sage: F = CombinatorialSpecies()
            sage: loads(dumps(F))
            Combinatorial species

        ::

            sage: X = species.SingletonSpecies()
            sage: E = species.EmptySetSpecies()
            sage: L = CombinatorialSpecies()
            sage: L.define(E+X*L)
            sage: L.generating_series()[0:4]
            [1, 1, 1, 1]
            sage: LL = loads(dumps(L))
            sage: LL.generating_series()[0:4]
            [1, 1, 1, 1]
        """
        self._generating_series = {}
        self._isotype_generating_series = {}
        self._cycle_index_series = {}
        GenericCombinatorialSpecies.__init__(self, min=min, max=None, weight=None)

    _default_structure_class = CombinatorialSpeciesStructure

    def __hash__(self):
        """
        EXAMPLES::

            sage: hash(CombinatorialSpecies) #random
            53751280

        ::

            sage: X = species.SingletonSpecies()
            sage: E = species.EmptySetSpecies()
            sage: L = CombinatorialSpecies()
            sage: L.define(E+X*L)
            sage: hash(L) #random
            -826511807095108317
        """
        try:
            return hash(('CombinatorialSpecies', id(self._reference)))
        except AttributeError:
            return hash('CombinatorialSpecies')

    def __eq__(self, other):
        """
        TESTS::

            sage: A = species.CombinatorialSpecies()
            sage: B = species.CombinatorialSpecies()
            sage: A == B
            False
            sage: X = species.SingletonSpecies()
            sage: A.define(X+A*A)
            sage: B.define(X+B*B)
            sage: A == B
            True

            sage: C = species.CombinatorialSpecies()
            sage: E = species.EmptySetSpecies()
            sage: C.define(E+X*C*C)
            sage: A == C
            False
        """
        if not isinstance(other, CombinatorialSpecies):
            return False
        if not hasattr(self, "_reference"):
            return False
        if hasattr(self, '_computing_eq'):
            return True

        self._computing_eq = True
        res = self._unique_info() == other._unique_info()
        del self._computing_eq
        return res

    def __ne__(self, other):
        """
        Check whether ``self`` is not equal to ``other``.

        EXAMPLES::

            sage: A = species.CombinatorialSpecies()
            sage: B = species.CombinatorialSpecies()
            sage: A != B
            True
            sage: X = species.SingletonSpecies()
            sage: A.define(X+A*A)
            sage: B.define(X+B*B)
            sage: A != B
            False

            sage: C = species.CombinatorialSpecies()
            sage: E = species.EmptySetSpecies()
            sage: C.define(E+X*C*C)
            sage: A != C
            True
        """
        return not (self == other)

    def _unique_info(self):
        """
        Return a tuple which should uniquely identify the species.

        EXAMPLES::

            sage: F = CombinatorialSpecies()
            sage: F._unique_info()
            (<class 'sage.combinat.species.recursive_species.CombinatorialSpecies'>,)

        ::

            sage: X = species.SingletonSpecies()
            sage: E = species.EmptySetSpecies()
            sage: L = CombinatorialSpecies()
            sage: L.define(E+X*L)
            sage: L._unique_info()
            (<class 'sage.combinat.species.recursive_species.CombinatorialSpecies'>,
             <class 'sage.combinat.species.sum_species.SumSpecies'>,
             None,
             None,
             1,
             Empty set species,
             Product of (Singleton species) and (Combinatorial species))
        """
        if hasattr(self, "_reference"):
            return (self.__class__,) + self._reference._unique_info()
        else:
            return (self.__class__,)

    def __getstate__(self):
        """
        EXAMPLES::

            sage: X = species.SingletonSpecies()
            sage: E = species.EmptySetSpecies()
            sage: L = CombinatorialSpecies()
            sage: L.define(E+X*L)
            sage: L.__getstate__()
            {'reference': Sum of (Empty set species) and (Product of (Singleton species) and (Combinatorial species))}
        """
        state = {}
        if hasattr(self, '_reference'):
            state['reference'] = self._reference
        return state

    def __setstate__(self, state):
        """
        EXAMPLES::

            sage: X = species.SingletonSpecies()
            sage: E = species.EmptySetSpecies()
            sage: L = CombinatorialSpecies()
            sage: L.define(E+X*L)
            sage: state = L.__getstate__(); state
            {'reference': Sum of (Empty set species) and (Product of (Singleton species) and (Combinatorial species))}
            sage: L._reference = None
            sage: L.__setstate__(state)
            sage: L._reference
            Sum of (Empty set species) and (Product of (Singleton species) and (Combinatorial species))
        """
        CombinatorialSpecies.__init__(self)
        if 'reference' in state:
            self.define(state['reference'])

    def _structures(self, structure_class, labels):
        """
        EXAMPLES::

            sage: F = CombinatorialSpecies()
            sage: list(F._structures(F._default_structure_class, [1,2,3]))
            Traceback (most recent call last):
            ...
            NotImplementedError
        """
        if not hasattr(self, "_reference"):
            raise NotImplementedError
        for s in self._reference.structures(labels):
            yield structure_class(self, s)

    def _isotypes(self, structure_class, labels):
        """
        EXAMPLES::

            sage: F = CombinatorialSpecies()
            sage: list(F._isotypes(F._default_structure_class, [1,2,3]))
            Traceback (most recent call last):
            ...
            NotImplementedError
        """
        if not hasattr(self, "_reference"):
            raise NotImplementedError
        for s in self._reference.isotypes(labels):
            yield structure_class(self, s)

    def _gs(self, series_ring, base_ring):
        """
        EXAMPLES::

            sage: F = CombinatorialSpecies()
            sage: F.generating_series()
            Uninitialized Lazy Series
        """
        if base_ring not in self._generating_series:
            self._generating_series[base_ring] = series_ring.undefined(valuation=(0 if self._min is None else self._min))

        res = self._generating_series[base_ring]
        if hasattr(self, "_reference") and not hasattr(res, "_reference"):
            res._reference = None
            res.define(self._reference.generating_series(base_ring))
        return res

    def _itgs(self, series_ring, base_ring):
        """
        EXAMPLES::

            sage: F = CombinatorialSpecies()
            sage: F.isotype_generating_series()
            Uninitialized Lazy Series
        """
        if base_ring not in self._isotype_generating_series:
            self._isotype_generating_series[base_ring] = series_ring.undefined(valuation=(0 if self._min is None else self._min))

        res = self._isotype_generating_series[base_ring]
        if hasattr(self, "_reference") and not hasattr(res, "_reference"):
            res._reference = None
            res.define(self._reference.isotype_generating_series(base_ring))
        return res

    def _cis(self, series_ring, base_ring):
        """
        EXAMPLES::

            sage: F = CombinatorialSpecies()
            sage: F.cycle_index_series()                                                # needs sage.modules
            Uninitialized Lazy Series
        """
        if base_ring not in self._cycle_index_series:
            self._cycle_index_series[base_ring] = series_ring.undefined(valuation=(0 if self._min is None else self._min))

        res = self._cycle_index_series[base_ring]
        if hasattr(self, "_reference") and not hasattr(res, "_reference"):
            res._reference = None
            res.define(self._reference.cycle_index_series(base_ring))
        return res

    def weight_ring(self):
        """
        EXAMPLES::

            sage: F = species.CombinatorialSpecies()
            sage: F.weight_ring()
            Rational Field

        ::

            sage: X = species.SingletonSpecies()
            sage: E = species.EmptySetSpecies()
            sage: L = CombinatorialSpecies()
            sage: L.define(E+X*L)
            sage: L.weight_ring()
            Rational Field
        """
        if not hasattr(self, "_reference"):
            return QQ

        if hasattr(self, "_weight_ring_been_called"):
            return QQ
        else:
            self._weight_ring_been_called = True
            res = self._reference.weight_ring()
            del self._weight_ring_been_called
            return res

    def define(self, x):
        """
        Define ``self`` to be equal to the combinatorial species ``x``.

        This is used to define combinatorial species recursively. All of the
        real work is done by calling the ``.set()`` method for each of the
        series associated to ``self``.

        EXAMPLES: The species of linear orders `L` can be recursively defined
        by `L = 1 + X*L` where 1 represents the empty set species
        and `X` represents the singleton species::

            sage: X = species.SingletonSpecies()
            sage: E = species.EmptySetSpecies()
            sage: L = CombinatorialSpecies()
            sage: L.define(E+X*L)
            sage: L.generating_series()[0:4]
            [1, 1, 1, 1]
            sage: L.structures([1,2,3]).cardinality()
            6
            sage: L.structures([1,2,3]).list()
            [1*(2*(3*{})),
             1*(3*(2*{})),
             2*(1*(3*{})),
             2*(3*(1*{})),
             3*(1*(2*{})),
             3*(2*(1*{}))]

        ::

            sage: L = species.LinearOrderSpecies()
            sage: L.generating_series()[0:4]
            [1, 1, 1, 1]
            sage: L.structures([1,2,3]).cardinality()
            6
            sage: L.structures([1,2,3]).list()
            [[1, 2, 3], [1, 3, 2], [2, 1, 3], [2, 3, 1], [3, 1, 2], [3, 2, 1]]

        TESTS::

            sage: A = CombinatorialSpecies()
            sage: A.define(E+X*A*A)
            sage: A.generating_series()[0:6]
            [1, 1, 2, 5, 14, 42]
            sage: A.generating_series().counts(6)
            [1, 1, 4, 30, 336, 5040]
            sage: len(A.structures([1,2,3,4]).list())
            336
            sage: A.isotype_generating_series()[0:6]
            [1, 1, 2, 5, 14, 42]
            sage: len(A.isotypes([1,2,3,4]).list())
            14

        ::

            sage: A = CombinatorialSpecies(min=1)
            sage: A.define(X+A*A)
            sage: A.generating_series()[0:6]
            [0, 1, 1, 2, 5, 14]
            sage: A.generating_series().counts(6)
            [0, 1, 2, 12, 120, 1680]
            sage: len(A.structures([1,2,3]).list())
            12
            sage: A.isotype_generating_series()[0:6]
            [0, 1, 1, 2, 5, 14]
            sage: len(A.isotypes([1,2,3,4]).list())
            5

        ::

            sage: X2 = X*X
            sage: X5 = X2*X2*X
            sage: A = CombinatorialSpecies(min=1)
            sage: B = CombinatorialSpecies(min=1)
            sage: C = CombinatorialSpecies(min=1)
            sage: A.define(X5+B*B)
            sage: B.define(X5+C*C)
            sage: C.define(X2+C*C+A*A)
            sage: A.generating_series()[0:15]
            [0, 0, 0, 0, 0, 1, 0, 0, 1, 2, 5, 4, 14, 10, 48]
            sage: B.generating_series()[0:15]
            [0, 0, 0, 0, 1, 1, 2, 0, 5, 0, 14, 0, 44, 0, 138]
            sage: C.generating_series()[0:15]
            [0, 0, 1, 0, 1, 0, 2, 0, 5, 0, 15, 0, 44, 2, 142]

        ::

            sage: F = CombinatorialSpecies()
            sage: F.define(E+X+(X*F+X*X*F))
            sage: F.generating_series().counts(10)
            [1, 2, 6, 30, 192, 1560, 15120, 171360, 2217600, 32296320]
            sage: F.generating_series()[0:10]
            [1, 2, 3, 5, 8, 13, 21, 34, 55, 89]
            sage: F.isotype_generating_series()[0:10]
            [1, 2, 3, 5, 8, 13, 21, 34, 55, 89]

        Check that :issue:`35071` is fixed::

            sage: X = species.SingletonSpecies()
            sage: E = species.SetSpecies(max=3)
            sage: B = species.CombinatorialSpecies(min=1)
            sage: B.define(X*E(B))
            sage: B.generating_series()
            z + z^2 + 3/2*z^3 + 5/2*z^4 + 9/2*z^5 + 17/2*z^6 + 133/8*z^7 + O(z^8)
        """
        if not isinstance(x, GenericCombinatorialSpecies):
            raise TypeError("x must be a combinatorial species")

        if self.__class__ is not CombinatorialSpecies:
            raise TypeError("only undefined combinatorial species can be set")

        self._reference = x

    def _add_to_digraph(self, d):
        """
        Add this species as a vertex to the digraph d along with any
        'children' of this species.

        Note that to avoid infinite recursion, we just return if this
        species already occurs in the digraph d.

        EXAMPLES::

            sage: d = DiGraph(multiedges=True)                                          # needs sage.graphs
            sage: X = species.SingletonSpecies()
            sage: B = species.CombinatorialSpecies()
            sage: B.define(X+B*B)
            sage: B._add_to_digraph(d); d                                               # needs sage.graphs
            Multi-digraph on 4 vertices

        TESTS::

            sage: C = species.CombinatorialSpecies()
            sage: C._add_to_digraph(d)                                                  # needs sage.graphs
            Traceback (most recent call last):
            ...
            NotImplementedError
        """
        if self in d:
            return
        try:
            d.add_edge(self, self._reference)
            self._reference._add_to_digraph(d)
        except AttributeError:
            raise NotImplementedError

    def _equation(self, var_mapping):
        """
        Return the right hand side of an algebraic equation satisfied by
        this species. This is a utility function called by the
        algebraic_equation_system method.

        EXAMPLES::

            sage: C = species.CombinatorialSpecies()
            sage: C.algebraic_equation_system()                                         # needs sage.graphs
            Traceback (most recent call last):
            ...
            NotImplementedError

        ::

            sage: B = species.BinaryTreeSpecies()
            sage: B.algebraic_equation_system()                                         # needs sage.graphs
            [-node3^2 + node1, -node1 + node3 + (-z)]
        """
        try:
            return var_mapping[self._reference]
        except AttributeError:
            raise NotImplementedError



# sage_setup: distribution = sagemath-combinat
r"""
Set of words

To define a new class of words, please refer to the documentation file:
sage/combinat/words/notes/word_inheritance_howto.rst

AUTHORS:

    - Franco Saliola (2008-12-17): merged into sage
    - Sebastien Labbe (2008-12-17): merged into sage
    - Arnaud Bergeron (2008-12-17): merged into sage
    - Sebastien Labbe (2009-07-21): Improved morphism iterator (:issue:`6571`).
    - Vincent Delecroix (2015): classes simplifications (:issue:`19619`)

EXAMPLES::

    sage: Words()
    Finite and infinite words over Set of Python objects of class 'object'
    sage: Words(4)
    Finite and infinite words over {1, 2, 3, 4}
    sage: Words(4,5)
    Words of length 5 over {1, 2, 3, 4}

    sage: FiniteWords('ab')
    Finite words over {'a', 'b'}
    sage: InfiniteWords('natural numbers')
    Infinite words over Non negative integers
"""
# ****************************************************************************
#       Copyright (C) 2008 Arnaud Bergeron <abergeron@gmail.com>,
#                          Sébastien Labbé <slabqc@gmail.com>,
#                          Franco Saliola <saliola@gmail.com>
#                     2015 Vincent Delecroix <20100.delecroix@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import itertools
from collections.abc import Iterable

from sage.misc.cachefunc import cached_method
from sage.misc.lazy_attribute import lazy_attribute
from sage.misc.persist import register_unpickle_override

from sage.structure.parent import Parent

from sage.categories.sets_cat import Sets

from sage.combinat.combinat import CombinatorialObject
from sage.structure.list_clone import ClonableElement
from sage.combinat.words.alphabet import build_alphabet

from sage.rings.infinity import Infinity
from sage.rings.integer import Integer
from sage.rings.integer_ring import ZZ


def Words(alphabet=None, length=None, finite=True, infinite=True):
    """
    Return the combinatorial class of words of length k over an alphabet.

    EXAMPLES::

        sage: Words()
        Finite and infinite words over Set of Python objects of class 'object'
        sage: Words(length=7)
        Words of length 7 over Set of Python objects of class 'object'
        sage: Words(5)
        Finite and infinite words over {1, 2, 3, 4, 5}
        sage: Words(5, 3)
        Words of length 3 over {1, 2, 3, 4, 5}
        sage: Words(5, infinite=False)
        Finite words over {1, 2, 3, 4, 5}
        sage: Words(5, finite=False)
        Infinite words over {1, 2, 3, 4, 5}
        sage: Words('ab')
        Finite and infinite words over {'a', 'b'}
        sage: Words('ab', 2)
        Words of length 2 over {'a', 'b'}
        sage: Words('ab', infinite=False)
        Finite words over {'a', 'b'}
        sage: Words('ab', finite=False)
        Infinite words over {'a', 'b'}
        sage: Words('positive integers', finite=False)
        Infinite words over Positive integers
        sage: Words('natural numbers')
        Finite and infinite words over Non negative integers
    """
    if isinstance(alphabet, (FiniteWords,
                             InfiniteWords,
                             FiniteOrInfiniteWords,
                             Words_n)):
        return alphabet

    if length is None:
        if finite and infinite:
            return FiniteOrInfiniteWords(alphabet)
        elif finite:
            return FiniteWords(alphabet)
        elif infinite:
            return InfiniteWords(alphabet)
        else:
            raise ValueError("either finite or infinite must be True")

    elif isinstance(length, (int, Integer)):
        return Words_n(FiniteWords(alphabet), length)

    raise ValueError("do not know how to make a combinatorial class of words from your input")


class AbstractLanguage(Parent):
    r"""
    Abstract base class.

    This is *not* to be used by any means. This class gather previous features
    of set of words (prior to :issue:`19619`). In the future that class might
    simply disappear or become a common base class for all languages. In the
    latter case, its name would possibly change to ``Language``.
    """
    def __init__(self, alphabet=None, category=None):
        r"""
        INPUT:

        - ``alphabet`` -- the underlying alphabet

        TESTS::

            sage: loads(dumps(FiniteWords('ab'))) == FiniteWords('ab')
            True
            sage: loads(dumps(InfiniteWords('ab'))) == InfiniteWords('ab')
            True

            sage: Words('abc').sortkey_letters
            <bound method AbstractLanguage._sortkey_trivial of ...>
            sage: Words('bac').sortkey_letters
            <bound method AbstractLanguage._sortkey_letters of ...>
        """
        if isinstance(alphabet, (int, Integer)):
            from sage.sets.integer_range import IntegerRange
            alphabet = IntegerRange(1, alphabet + 1)
        elif (alphabet == "integers" or
              alphabet == "positive integers" or
              alphabet == "natural numbers"):
            alphabet = build_alphabet(name=alphabet)
        else:
            alphabet = build_alphabet(alphabet)

        self._alphabet = alphabet

        # Default sorting key: use rank()
        self.sortkey_letters = self._sortkey_letters

        # Check if we should use the trivial sorting key
        N = alphabet.cardinality()
        if N == Infinity:
            self.sortkey_letters = self._sortkey_trivial
        elif N < 36:
            try:
                if all(alphabet.unrank(i) > alphabet.unrank(j)
                       for i in range(N) for j in range(i)):
                    self.sortkey_letters = self._sortkey_trivial
            except TypeError:
                pass

        if category is None:
            category = Sets()

        Parent.__init__(self, category=category)

    def alphabet(self):
        r"""
        EXAMPLES::

            sage: Words(NN).alphabet()
            Non negative integer semiring

            sage: InfiniteWords([1,2,3]).alphabet()
            {1, 2, 3}
            sage: InfiniteWords('ab').alphabet()
            {'a', 'b'}

            sage: FiniteWords([1,2,3]).alphabet()
            {1, 2, 3}
            sage: FiniteWords().alphabet()
            Set of Python objects of class 'object'
        """
        return self._alphabet

    def identity_morphism(self):
        r"""
        Return the identity morphism from ``self`` to itself.

        EXAMPLES::

            sage: W = Words('ab')
            sage: W.identity_morphism()
            WordMorphism: a->a, b->b

        ::

            sage: W = Words(range(3))
            sage: W.identity_morphism()
            WordMorphism: 0->0, 1->1, 2->2

        There is no support yet for infinite alphabet::

            sage: W = Words(alphabet=Alphabet(name='NN'))
            sage: W
            Finite and infinite words over Non negative integers
            sage: W.identity_morphism()
            Traceback (most recent call last):
            ...
            NotImplementedError: size of alphabet must be finite
        """
        if self.alphabet().cardinality() not in ZZ:
            raise NotImplementedError('size of alphabet must be finite')
        from sage.combinat.words.morphism import WordMorphism
        return WordMorphism({a: a for a in self.alphabet()})

    def _check(self, w, length=40):
        r"""
        Check that the first length elements are actually in the alphabet.

        INPUT:

        - ``w`` -- word

        - ``length`` -- integer (default: `40`)

        EXAMPLES::

            sage: W = FiniteWords(['a','b','c'])
            sage: W._check('abcabc') is None
            True
            sage: W._check('abcabcd')
            Traceback (most recent call last):
            ...
            ValueError: d not in alphabet
            sage: W._check('abcabc'*10+'z') is None
            True
            sage: W._check('abcabc'*10+'z', length=80)
            Traceback (most recent call last):
            ...
            ValueError: z not in alphabet
        """
        stop = None if length is None else int(length)
        for a in itertools.islice(w, stop):
            if a not in self.alphabet():
                raise ValueError("%s not in alphabet" % a)

    def _sortkey_trivial(self, letter1):
        """
        Trivial function, used to sort the letters by their names.

        INPUT:

        - ``letter1`` -- a letter in the alphabet

        EXAMPLES::

            sage: W = FiniteWords('ade')
            sage: W.sortkey_letters('d')  # indirect doctest
            'd'
        """
        return letter1

    def _sortkey_letters(self, letter1):
        r"""
        Return the default value used to sort the letters.

        INPUT:

        - ``letter1`` -- a letter in the alphabet

        EXAMPLES::

            sage: W = FiniteWords('woa')
            sage: W.sortkey_letters('w')  # indirect doctest
            0
            sage: W.sortkey_letters('o')  # indirect doctest
            1
            sage: W.sortkey_letters('a')  # indirect doctest
            2

        TESTS::

            sage: assert W.sortkey_letters == W._sortkey_letters
        """
        rk = self.alphabet().rank
        return rk(letter1)

    def __eq__(self, other) -> bool:
        r"""
        TESTS::

            sage: FiniteWords() == FiniteWords()
            True
            sage: FiniteWords() == InfiniteWords()
            False
            sage: InfiniteWords() == Words()
            False
            sage: FiniteWords([0,1]) == FiniteWords([0,1,2,3])
            False
        """
        return self is other or (type(self) is type(other) and
                                 self.alphabet() == other.alphabet())

    def __ne__(self, other) -> bool:
        r"""
        TESTS::

            sage: InfiniteWords() != InfiniteWords()
            False
            sage: FiniteWords() != Words()
            True
            sage: Words('ab') != Words('ab')
            False
        """
        return not (self == other)


class FiniteWords(AbstractLanguage):
    r"""
    The set of finite words over a fixed alphabet.

    EXAMPLES::

        sage: W = FiniteWords('ab')
        sage: W
        Finite words over {'a', 'b'}

    TESTS::

        sage: TestSuite(FiniteWords('ab')).run()
        sage: TestSuite(FiniteWords([])).run()
        sage: TestSuite(FiniteWords(['a'])).run()
    """

    def __init__(self, alphabet=None, category=None):
        """
        INPUT:

        - ``alphabet`` -- the underlying alphabet
        - ``category`` -- the suggested category of the set
          (normally should be automatically determined)

        TESTS::

            sage: FiniteWords('ab').is_finite()
            False
            sage: FiniteWords('ab').category()
            Category of infinite sets
            sage: FiniteWords([]).is_finite()
            True
            sage: FiniteWords([]).category()
            Category of finite sets
            sage: FiniteWords([], Sets()).category()
            Category of finite sets
        """
        if category is None:
            category = Sets()
        if alphabet:
            category = category.Infinite()
        else:
            category = category.Finite()
        super().__init__(alphabet, category)

    def is_empty(self):
        """
        Return ``False``, because the empty word is in the set.

        TESTS::

            sage: FiniteWords('ab').is_empty()
            False
            sage: FiniteWords([]).is_empty()
            False
        """
        return False

    def cardinality(self):
        r"""
        Return the cardinality of this set.

        EXAMPLES::

            sage: FiniteWords('').cardinality()
            1
            sage: FiniteWords('a').cardinality()
            +Infinity
        """
        if not self.alphabet():
            return ZZ.one()
        return Infinity

    def __hash__(self):
        r"""
        TESTS::

            sage: hash(FiniteWords('ab')) # random
            12
        """
        return hash(self.alphabet()) ^ hash('finite words')

    @cached_method
    def shift(self):
        r"""
        Return the set of infinite words on the same alphabet.

        EXAMPLES::

            sage: FiniteWords('ab').shift()
            Infinite words over {'a', 'b'}
        """
        return InfiniteWords(self.alphabet())

    def factors(self):
        r"""
        Return itself.

        EXAMPLES::

            sage: FiniteWords('ab').factors()
            Finite words over {'a', 'b'}
        """
        return self

    @lazy_attribute
    def _element_classes(self):
        r"""
        Return a dictionary that gives the class of the element of ``self``.

        The word may be finite, infinite or of unknown length.
        Its data may be string, list, tuple, a callable or an iterable.
        For callable and iterable, the data may be cached.

        EXAMPLES:

        Once you get the class, it can be used to create a new word::

            sage: W = FiniteWords([0,1,2])
            sage: L = [0,1,0] * 100
            sage: cls = W._element_classes['list']
            sage: w = cls(W, L)
            sage: type(w)
            <class 'sage.combinat.words.word.FiniteWord_list'>
            sage: w
            word: 0100100100100100100100100100100100100100...
            sage: w.parent()
            Finite words over {0, 1, 2}

        TESTS::

            sage: d = FiniteWords()._element_classes
            sage: type(d)
            <class 'dict'>
            sage: len(d)
            7
            sage: e = FiniteWords('abcdefg')._element_classes
            sage: d == e
            True
        """
        import sage.combinat.words.word as word
        classes = {
            'list': word.FiniteWord_list,
            'str': word.FiniteWord_str,
            'tuple': word.FiniteWord_tuple,
            'callable_with_caching': word.FiniteWord_callable_with_caching,
            'callable': word.FiniteWord_callable,
            'iter_with_caching': word.FiniteWord_iter_with_caching,
            'iter': word.FiniteWord_iter}

        # test whether or not we can use the class Finiteword_char
        if (self.alphabet().cardinality() <= 256 and
                all(isinstance(i, (int, Integer)) and
                    0 <= i < 256 for i in self.alphabet())):
            L = self.alphabet().list()
            key = self.sortkey_letters
            if (all(L[i] < L[i + 1] for i in range(len(L) - 1)) and
                    all(key(L[i]) < key(L[i + 1]) for i in range(len(L) - 1))):
                classes['char'] = word.FiniteWord_char

        return classes

    def _word_from_word(self, data):
        r"""
        Return a word from a word.

        The data is assumed to be ok, no check is performed.

        INPUT:

        - ``data`` -- word

        EXAMPLES::

            sage: W = FiniteWords([0,1,2])
            sage: w = W([0,1,2,0,1,2])
            sage: z = W._word_from_word(w)
            sage: z
            word: 012012
            sage: w is z
            True
        """
        ####################
        # If `data` is already a word and if its parent is self, then
        # return `data`.
        ###########################
        if data.parent() is self or data.parent() == self:
            return data

        ###########################
        # Otherwise, if self is not the parent of `data`, then we try to
        # recover the data, the length and the datatype of the input `data`
        # To minimize the impact of the import, we do it only at the time there
        # are needed
        ###########################
        from sage.combinat.words.word_char import WordDatatype_char
        if isinstance(data, WordDatatype_char):
            data = list(data)
            if 'char' in self._element_classes:
                return self._element_classes['char'](self, data)
            else:
                return self._element_classes['list'](self, data)

        from sage.combinat.words.word_datatypes import (WordDatatype_str,
                                                        WordDatatype_list,
                                                        WordDatatype_tuple)
        if isinstance(data, WordDatatype_str):
            return self._element_classes['str'](self, data._data)
        if isinstance(data, WordDatatype_tuple):
            return self._element_classes['tuple'](self, data._data)
        if isinstance(data, WordDatatype_list):
            return self._element_classes['list'](self, data._data)

        from sage.combinat.words.word_infinite_datatypes import \
            (WordDatatype_callable, WordDatatype_iter)
        if isinstance(data, WordDatatype_callable):
            length = data.length()
            data = data._func
            return self._word_from_callable(data, length, caching=False)
        if isinstance(data, WordDatatype_iter):
            length = data.length()
            data = iter(data)
            return self._word_from_iter(data, length, caching=False)

        raise TypeError("any instance of Word_class must be an instance of WordDatatype")

    def _word_from_callable(self, data, length, caching=True):
        r"""
        Return a word represented by a callable.

        The data is assumed to be ok, no check is performed.

        INPUT:

        - ``data`` -- callable
        - ``length`` -- integer or ``None`` or ``'infinite'`` or ``Infinity``
        - ``caching`` -- boolean (default: ``True``); whether to keep a cache
          of the letters computed by the callable

        EXAMPLES::

            sage: W = FiniteWords([0,1,2])
            sage: f = lambda n : n % 3
            sage: W._word_from_callable(f, 100)
            word: 0120120120120120120120120120120120120120...
        """
        wc = '_with_caching' if caching else ""
        if length not in ZZ or length < 0:
            raise ValueError("not a correct value for length (%s)" % length)
        return self._element_classes['callable' + wc](self, data, length)

    def _word_from_iter(self, data, length=None, caching=True):
        r"""
        Return a word represented by an iterator.

        The data is assumed to be ok, no check is performed.

        INPUT:

        - ``data`` -- iterable

        - ``length`` -- (optional) integer

        - ``caching`` -- boolean (default: ``True``); whether to keep a cache
          of the letters computed by the iterator

        EXAMPLES::

            sage: W = FiniteWords([0,1,2])
            sage: W._word_from_iter(iter([1]*10))
            word: 1111111111
            sage: W._word_from_iter(iter([1]*10), 5)
            word: 11111
        """
        wc = '_with_caching' if caching else ""
        if length is None or length == "finite":
            length = "finite"
        elif length not in ZZ or length < 0:
            raise ValueError("not a correct value for length (%s)" % length)
        return self._element_classes['iter' + wc](self, data, length)

    def __call__(self, data=None, length=None, datatype=None, caching=True, check=True):
        r"""
        Construct a new word object with parent ``self``.

        INPUT:

        - ``data`` -- (default: ``None``) list, string, tuple, iterator, ``None``
          (shorthand for []), or a callable defined on [0,1,...,length]

        - ``length`` -- integer (default: ``None``); only used if the data is
          an iterator or a callable. It determines the length of the word

        - ``datatype`` -- (default: ``None``) ``None``, "char", "list", "str",
          "tuple", "iter", "callable" or "pickled_function"; if ``None``, then
          the function tries to guess this from the data

        - ``caching`` -- boolean (default: ``True``); whether to keep a cache
          of the letters computed by an iterator or callable

        - ``check`` -- boolean (default: ``True``); whether to check if
          the 40 first letters are in the parent alphabet. This is a
          check done to test for small programming errors. Since we also
          support infinite words, we cannot really implement a more
          accurate check.

        .. NOTE::

           The check makes this method about 10 times slower (20µs instead
           of 2µs), so make sure to set it to False if you know the
           alphabet is OK. Fast creation (about 1µs) of a word can be
           done using the class directly (see :meth:`_element_classes`).

        .. WARNING::

           Be careful when defining words using callables and iterators. It
           appears that islice does not pickle correctly causing various errors
           when reloading. Also, most iterators do not support copying and
           should not support pickling by extension.

        EXAMPLES::

            sage: W = FiniteWords()

        Empty word::

            sage: W()
            word:

        Word with string::

            sage: W("abbabaab")
            word: abbabaab

        Word with string constructed from other types::

            sage: W([0,1,1,0,1,0,0,1], datatype='str')
            word: 01101001
            sage: W((0,1,1,0,1,0,0,1), datatype='str')
            word: 01101001

        Word with list::

            sage: W([0,1,1,0,1,0,0,1])
            word: 01101001

        Word with list constructed from other types::

            sage: W("01101001", datatype='list')
            word: 01101001
            sage: W((0,1,1,0,1,0,0,1), datatype='list')
            word: 01101001

        Word with tuple::

            sage: W((0,1,1,0,1,0,0,1))
            word: 01101001

        Word with tuple constructed from other types::

            sage: W([0,1,1,0,1,0,0,1], datatype='tuple')
            word: 01101001
            sage: W("01101001", datatype='str')
            word: 01101001

        Word with iterator::

            sage: from itertools import count
            sage: W(count(), length=10)
            word: 0123456789
            sage: W(iter("abbabaab"))
            word: abbabaab

        Word with function (a 'callable')::

            sage: f = lambda n : add(Integer(n).digits(2)) % 2
            sage: W(f, length=12)
            word: 011010011001
            sage: FiniteWords([0,1,2])(f, length=12)
            word: 011010011001

        Word over a string with a parent::

            sage: w = FiniteWords('abc')("abbabaab"); w
            word: abbabaab
            sage: w.parent()
            Finite words over {'a', 'b', 'c'}

        The fourty first letters of the word are checked if they are in the
        parent alphabet::

            sage: FiniteWords("ab")("abca")
            Traceback (most recent call last):
            ...
            ValueError: c not in alphabet
            sage: FiniteWords("ab")("abca", check=False)
            word: abca

        The default parent is the combinatorial class of all words::

            sage: w = Word("abbabaab"); w
            word: abbabaab
            sage: w.parent()
            Finite words over Set of Python objects of class 'object'

        Creation of a word from a word::

            sage: FiniteWords([0,1,2,3])(FiniteWords([2,3])([2,2,2,3,3,2]))
            word: 222332
            sage: _.parent()
            Finite words over {0, 1, 2, 3}

        ::

            sage: FiniteWords([3,2,1])(FiniteWords([2,3])([2,2,2,3,3,2]))
            word: 222332
            sage: _.parent()
            Finite words over {3, 2, 1}

        Construction of a word from a word when the parents are the same::

            sage: W = FiniteWords()
            sage: w = W(range(8))
            sage: z = W(w)
            sage: w is z
            True

        Construction of a word path from a finite word::

            sage: W = FiniteWords('abcd')
            sage: P = WordPaths('abcd')                                                 # needs sage.modules
            sage: w = W('aaab')
            sage: P(w)                                                                  # needs sage.modules
            Path: aaab

        Construction of a word path from a Christoffel word::

            sage: w = words.ChristoffelWord(5,8)
            sage: w
            word: 0010010100101
            sage: P = WordPaths([0,1,2,3])                                              # needs sage.modules
            sage: P(w)                                                                  # needs sage.modules
            Path: 0010010100101

        Construction of a word represented by a list from a word
        represented by a str ::

            sage: w = W('ababbbabab')
            sage: type(w)
            <class 'sage.combinat.words.word.FiniteWord_str'>
            sage: z = Word(w, datatype='list')
            sage: type(z)
            <class 'sage.combinat.words.word.FiniteWord_list'>
            sage: y = Word(w, alphabet='abc', datatype='list')
            sage: type(y)
            <class 'sage.combinat.words.word.FiniteWord_list'>

        Creation of a word from a concatenation of words::

            sage: W = FiniteWords()
            sage: w = W() * W('a')
            sage: Z = Words('ab')
            sage: Z(w)
            word: a

        Creation of a word path from a FiniteWord_iter::

            sage: w = words.FibonacciWord()
            sage: f = w[:100]
            sage: P = WordPaths([0,1,2,3])                                              # needs sage.modules
            sage: p = P(f); p                                                           # needs sage.modules
            Path: 0100101001001010010100100101001001010010...
            sage: p.length()                                                            # needs sage.modules
            100

        Creation of a word path from a FiniteWord_callable::

            sage: g = W(lambda n:n%2, length = 100)
            sage: P = WordPaths([0,1,2,3])                                              # needs sage.modules
            sage: p = P(g); p                                                           # needs sage.modules
            Path: 0101010101010101010101010101010101010101...
            sage: p.length()                                                            # needs sage.modules
            100

        Creation of a word from a pickled function::

            sage: f = lambda n : n % 10
            sage: from sage.misc.fpickle import pickle_function
            sage: s = pickle_function(f)
            sage: W(s, length=10, datatype='pickled_function')
            word: 0123456789

        If the alphabet is a subset of [0, 255], then it uses char as datatype::

            sage: type(Word([0,1,1,2,0], alphabet=list(range(256))))
            <class 'sage.combinat.words.word.FiniteWord_char'>

        If the alphabet is a subset of [0, 255], then the letters must
        convert to an unsigned char. Otherwise an error is raised before
        the check is done::

            sage: type(Word([0,1,1,2,0,257], alphabet=list(range(256))))
            Traceback (most recent call last):
            ...
            OverflowError: value too large to convert to unsigned char
            sage: type(Word([0,1,1,2,0,258], alphabet=list(range(257))))
            Traceback (most recent call last):
            ...
            ValueError: 258 not in alphabet
            sage: type(Word([0,1,1,2,0,103], alphabet=list(range(100))))
            Traceback (most recent call last):
            ...
            ValueError: 103 not in alphabet
        """
        if datatype is not None:
            if datatype == 'list':
                w = self._element_classes['list'](self, data)
            elif datatype == 'char':
                w = self._element_classes['char'](self, data)
            elif datatype == 'tuple':
                w = self._element_classes['tuple'](self, data)
            elif datatype == 'str':
                w = self._element_classes['str'](self, data)
            elif datatype == 'callable':
                w = self._word_from_callable(data, length, caching)
            elif datatype == 'iter':
                w = self._word_from_iter(data, length, caching)
            elif datatype == 'pickled_function':
                from sage.misc.fpickle import unpickle_function
                data = unpickle_function(data)
                w = self._word_from_callable(data, length, caching)
            else:
                raise ValueError("unknown datatype (={})".format(datatype))

        elif 'char' in self._element_classes:
            if data is None:
                data = []
            elif callable(data):
                data = [data(i) for i in range(length)]
            elif not isinstance(data, (tuple, list)):
                data = list(data)
            w = self._element_classes['char'](self, data)

        elif isinstance(data, list):
            w = self._element_classes['list'](self, data)

        elif data is None:
            w = self._element_classes['list'](self, [])

        elif isinstance(data, str):
            w = self._element_classes['str'](self, data)

        elif isinstance(data, tuple):
            w = self._element_classes['tuple'](self, data)

        elif isinstance(data, (CombinatorialObject, ClonableElement)):
            w = self._element_classes['list'](self, list(data))

        elif callable(data):
            w = self._word_from_callable(data, length, caching)

        elif isinstance(data, Iterable):
            from sage.combinat.words.abstract_word import Word_class
            if isinstance(data, Word_class):
                w = self._word_from_word(data)
            else:
                w = self._word_from_iter(data, length, caching)

        else:
            raise ValueError("cannot guess a datatype from data (=%s); please specify one" % data)

        if check:
            self._check(w)
        return w

    def _repr_(self) -> str:
        """
        EXAMPLES::

            sage: FiniteWords() # indirect doctest
            Finite words over Set of Python objects of class 'object'
        """
        return 'Finite words over {!r}'.format(self.alphabet())

    def _an_element_(self):
        r"""
        Return an element of ``self``.

        EXAMPLES::

            sage: FiniteWords(4).an_element() # indirect doctest
            word: 212
            sage: FiniteWords([5, 1, 9]).an_element() # indirect doctest
            word: 151
            sage: FiniteWords([1]).an_element() # indirect doctest
            word: 111
            sage: FiniteWords(NN).an_element() # indirect doctest
            word: 101
        """
        try:
            some_letters = list(self.alphabet().some_elements())
        except (TypeError, ValueError, AttributeError, NotImplementedError):
            return self([])

        if len(some_letters) == 0:
            return self([])
        if len(some_letters) == 1:
            return self([some_letters[0]] * 3)

        a, b = some_letters[:2]
        return self([b, a, b])

    def iterate_by_length(self, l=1):
        r"""
        Return an iterator over all the words of ``self`` of length `l`.

        INPUT:

        - ``l`` -- integer (default: 1); the length of the desired words

        EXAMPLES::

            sage: W = FiniteWords('ab')
            sage: list(W.iterate_by_length(1))
            [word: a, word: b]
            sage: list(W.iterate_by_length(2))
            [word: aa, word: ab, word: ba, word: bb]
            sage: list(W.iterate_by_length(3))
            [word: aaa,
             word: aab,
             word: aba,
             word: abb,
             word: baa,
             word: bab,
             word: bba,
             word: bbb]
            sage: list(W.iterate_by_length('a'))
            Traceback (most recent call last):
            ...
            TypeError: the parameter l (='a') must be an integer

        TESTS::

            sage: W = FiniteWords(NN)
            sage: list(W.iterate_by_length(1))
            Traceback (most recent call last):
            ...
            NotImplementedError: cannot iterate over words for infinite alphabets
        """
        if not isinstance(l, (int, Integer)):
            raise TypeError("the parameter l (=%r) must be an integer" % l)
        cls = self._element_classes['tuple']
        if not self.alphabet().is_finite():
            raise NotImplementedError("cannot iterate over words for infinite alphabets")
        for w in itertools.product(self.alphabet(), repeat=l):
            yield cls(self, w)

    def __iter__(self):
        r"""
        Return an iterator over all the words of ``self``.

        The iterator outputs the words in shortlex order (see
        :wikipedia:`Shortlex_order`), i.e. first by increasing length and then
        lexicographically.

        EXAMPLES::

            sage: W = Words([4,5], infinite=False)
            sage: for w in W:
            ....:   if len(w)>3:
            ....:       break
            ....:   else:
            ....:       w
            word:
            word: 4
            word: 5
            word: 44
            word: 45
            ...
            word: 554
            word: 555
            sage: W = Words([5,4], infinite=False)
            sage: for w in W:
            ....:   if len(w)>3:
            ....:       break
            ....:   else:
            ....:       w
            word:
            word: 5
            word: 4
            word: 55
            word: 54
            ...
            word: 445
            word: 444
        """
        for l in itertools.count():
            yield from self.iterate_by_length(l)

    def __contains__(self, x) -> bool:
        """
        Test whether ``self`` contains ``x``.

        OUTPUT:

            This method returns ``True`` if ``x`` is a word of the appropriate
            length and the alphabets of the parents match. It returns ``False``
            otherwise.

        EXAMPLES::

            sage: W = FiniteWords('ab')
            sage: W('ab') in W
            True
            sage: W('aa') in FiniteWords('aa')
            False
            sage: FiniteWords('a')('aa') in FiniteWords('ab')
            False
            sage: 2 in FiniteWords([1,2,3])
            False
            sage: [2] in FiniteWords([1,2,3])
            False
            sage: [1, 'a'] in FiniteWords([1,2,3])
            False
        """
        from sage.combinat.words.finite_word import FiniteWord_class
        return isinstance(x, FiniteWord_class) and x.parent().alphabet() == self.alphabet()

    def random_element(self, length=None, *args, **kwds):
        r"""
        Return a random finite word on the given alphabet.

        INPUT:

        - ``length`` -- (optional) the length of the word; if not set, will use
          a uniformly random number between 0 and 10

        - all other argument are transmitted to the random generator of the
          alphabet

        EXAMPLES::

            sage: W = FiniteWords(5)
            sage: W.random_element() # random
            word: 5114325445423521544531411434451152142155...

            sage: W = FiniteWords(ZZ)
            sage: W.random_element() # random
            word: 5,-1,-1,-1,0,0,0,0,-3,-11
            sage: W.random_element(length=4, x=0, y=4) # random
            word: 1003

        TESTS::

            sage: _ = FiniteWords(GF(5)).random_element()                               # needs sage.rings.finite_rings
        """
        if length is None:
            length = ZZ.random_element(0, 10)
        return self([self.alphabet().random_element(*args, **kwds)
                     for x in range(length)])

    def iter_morphisms(self, arg=None, codomain=None, min_length=1):
        r"""
        Iterate over all morphisms with domain ``self`` and the given
        codomain.

        INPUT:

        - ``arg`` -- (default: ``None``) it can be one of the following:

          - ``None`` -- then the method iterates through all morphisms

          - tuple `(a, b)` of two integers -- it specifies the range
            ``range(a, b)`` of values to consider for the sum of the length
            of the image of each letter in the alphabet

          - list of nonnegative integers -- the length of the list must be
            equal to the size of the alphabet, and the `i`-th integer of
            ``arg`` determines the length of the word mapped to by the `i`-th
            letter of the (ordered) alphabet

        - ``codomain`` -- (default: ``None``) a combinatorial class of words;
          by default, ``codomain`` is ``self``

        - ``min_length`` -- nonnegative integer (default: 1); if ``arg`` is
          not specified, then iterate through all the morphisms where the
          length of the images of each letter in the alphabet is at least
          ``min_length``. This is ignored if ``arg`` is a list.

        OUTPUT: iterator

        EXAMPLES:

        Iterator over all non-erasing morphisms::

            sage: W = FiniteWords('ab')
            sage: it = W.iter_morphisms()
            sage: for _ in range(7): next(it)
            WordMorphism: a->a, b->a
            WordMorphism: a->a, b->b
            WordMorphism: a->b, b->a
            WordMorphism: a->b, b->b
            WordMorphism: a->aa, b->a
            WordMorphism: a->aa, b->b
            WordMorphism: a->ab, b->a

        Iterator over all morphisms including erasing morphisms::

            sage: W = FiniteWords('ab')
            sage: it = W.iter_morphisms(min_length=0)
            sage: for _ in range(7): next(it)
            WordMorphism: a->, b->
            WordMorphism: a->a, b->
            WordMorphism: a->b, b->
            WordMorphism: a->, b->a
            WordMorphism: a->, b->b
            WordMorphism: a->aa, b->
            WordMorphism: a->ab, b->

        Iterator over morphisms where the sum of the lengths of the images
        of the letters is in a specific range::

            sage: for m in W.iter_morphisms((0, 3), min_length=0): m
            WordMorphism: a->aa, b->
            WordMorphism: a->ab, b->
            WordMorphism: a->ba, b->
            WordMorphism: a->bb, b->
            WordMorphism: a->a, b->a
            WordMorphism: a->a, b->b
            WordMorphism: a->b, b->a
            WordMorphism: a->b, b->b
            WordMorphism: a->a, b->
            WordMorphism: a->b, b->
            WordMorphism: a->, b->aa
            WordMorphism: a->, b->ab
            WordMorphism: a->, b->ba
            WordMorphism: a->, b->bb
            WordMorphism: a->, b->a
            WordMorphism: a->, b->b
            WordMorphism: a->, b->

        ::

            sage: for m in W.iter_morphisms( (2, 4) ): m
            WordMorphism: a->aa, b->a
            WordMorphism: a->aa, b->b
            WordMorphism: a->ab, b->a
            WordMorphism: a->ab, b->b
            WordMorphism: a->ba, b->a
            WordMorphism: a->ba, b->b
            WordMorphism: a->bb, b->a
            WordMorphism: a->bb, b->b
            WordMorphism: a->a, b->aa
            WordMorphism: a->a, b->ab
            WordMorphism: a->a, b->ba
            WordMorphism: a->a, b->bb
            WordMorphism: a->b, b->aa
            WordMorphism: a->b, b->ab
            WordMorphism: a->b, b->ba
            WordMorphism: a->b, b->bb
            WordMorphism: a->a, b->a
            WordMorphism: a->a, b->b
            WordMorphism: a->b, b->a
            WordMorphism: a->b, b->b

        Iterator over morphisms with specific image lengths::

            sage: for m in W.iter_morphisms([0, 0]): m
            WordMorphism: a->, b->
            sage: for m in W.iter_morphisms([0, 1]): m
            WordMorphism: a->, b->a
            WordMorphism: a->, b->b
            sage: for m in W.iter_morphisms([2, 1]): m
            WordMorphism: a->aa, b->a
            WordMorphism: a->aa, b->b
            WordMorphism: a->ab, b->a
            WordMorphism: a->ab, b->b
            WordMorphism: a->ba, b->a
            WordMorphism: a->ba, b->b
            WordMorphism: a->bb, b->a
            WordMorphism: a->bb, b->b
            sage: for m in W.iter_morphisms([2, 2]): m
            WordMorphism: a->aa, b->aa
            WordMorphism: a->aa, b->ab
            WordMorphism: a->aa, b->ba
            WordMorphism: a->aa, b->bb
            WordMorphism: a->ab, b->aa
            WordMorphism: a->ab, b->ab
            WordMorphism: a->ab, b->ba
            WordMorphism: a->ab, b->bb
            WordMorphism: a->ba, b->aa
            WordMorphism: a->ba, b->ab
            WordMorphism: a->ba, b->ba
            WordMorphism: a->ba, b->bb
            WordMorphism: a->bb, b->aa
            WordMorphism: a->bb, b->ab
            WordMorphism: a->bb, b->ba
            WordMorphism: a->bb, b->bb

        The codomain may be specified as well::

            sage: Y = FiniteWords('xyz')
            sage: for m in W.iter_morphisms([0, 2], codomain=Y): m
            WordMorphism: a->, b->xx
            WordMorphism: a->, b->xy
            WordMorphism: a->, b->xz
            WordMorphism: a->, b->yx
            WordMorphism: a->, b->yy
            WordMorphism: a->, b->yz
            WordMorphism: a->, b->zx
            WordMorphism: a->, b->zy
            WordMorphism: a->, b->zz
            sage: for m in Y.iter_morphisms([0,2,1], codomain=W): m
            WordMorphism: x->, y->aa, z->a
            WordMorphism: x->, y->aa, z->b
            WordMorphism: x->, y->ab, z->a
            WordMorphism: x->, y->ab, z->b
            WordMorphism: x->, y->ba, z->a
            WordMorphism: x->, y->ba, z->b
            WordMorphism: x->, y->bb, z->a
            WordMorphism: x->, y->bb, z->b
            sage: it = W.iter_morphisms(codomain=Y)
            sage: for _ in range(10): next(it)
            WordMorphism: a->x, b->x
            WordMorphism: a->x, b->y
            WordMorphism: a->x, b->z
            WordMorphism: a->y, b->x
            WordMorphism: a->y, b->y
            WordMorphism: a->y, b->z
            WordMorphism: a->z, b->x
            WordMorphism: a->z, b->y
            WordMorphism: a->z, b->z
            WordMorphism: a->xx, b->x

        TESTS::

            sage: list(W.iter_morphisms([1,0]))
            [WordMorphism: a->a, b->, WordMorphism: a->b, b->]
            sage: list(W.iter_morphisms([0,0], codomain=Y))
            [WordMorphism: a->, b->]
            sage: list(W.iter_morphisms([0, 1, 2]))
            Traceback (most recent call last):
            ...
            TypeError: arg (=[0, 1, 2]) must be an iterable of 2 integers
            sage: list(W.iter_morphisms([0, 'a']))
            Traceback (most recent call last):
            ...
            TypeError: arg (=[0, 'a']) must be an iterable of 2 integers
            sage: list(W.iter_morphisms([0, 1], codomain='a'))
            Traceback (most recent call last):
            ...
            TypeError: codomain (=a) must be an instance of FiniteWords
        """
        n = self.alphabet().cardinality()
        min_length = max(min_length, 0)
        # create an iterable of compositions (all "compositions" if arg is
        # None, or [arg] otherwise)
        if arg is None:
            from sage.combinat.integer_lists.nn import IntegerListsNN
            compositions = IntegerListsNN(length=n, min_part=min_length)
        elif isinstance(arg, tuple):
            from sage.combinat.integer_lists import IntegerListsLex
            a, b = arg
            compositions = IntegerListsLex(min_sum=a, max_sum=b - 1,
                                           length=n, min_part=min_length)
        else:
            arg = list(arg)
            if (not len(arg) == n or not
                    all(isinstance(a, (int, Integer)) for a in arg)):
                raise TypeError(
                    "arg (=%s) must be an iterable of %s integers" % (arg, n))
            compositions = [arg]

        # set the codomain
        if codomain is None:
            codomain = self
        elif isinstance(codomain, FiniteOrInfiniteWords):
            codomain = codomain.finite_words()
        elif not isinstance(codomain, FiniteWords):
            raise TypeError("codomain (=%s) must be an instance of FiniteWords" % codomain)

        # iterate through the morphisms
        from sage.combinat.words.morphism import WordMorphism
        for composition in compositions:
            cuts = [0] + list(composition)
            for i in range(1, len(cuts)):
                cuts[i] += cuts[i - 1]
            s = cuts[-1]  # same but better than s = sum(composition)
            for big_word in codomain.iterate_by_length(s):
                d = {}
                i = 0
                for a in self.alphabet():
                    d[a] = big_word[cuts[i]:cuts[i + 1]]
                    i += 1
                yield WordMorphism(d, codomain=codomain)


class InfiniteWords(AbstractLanguage):
    def cardinality(self):
        r"""
        Return the cardinality of this set.

        EXAMPLES::

            sage: InfiniteWords('ab').cardinality()
            +Infinity
            sage: InfiniteWords('a').cardinality()
            1
            sage: InfiniteWords('').cardinality()
            0
        """
        if not self.alphabet().cardinality():
            return ZZ.zero()
        elif self.alphabet().cardinality().is_one():
            return ZZ.one()
        else:
            return Infinity

    def __hash__(self):
        r"""
        TESTS::

            sage: hash(InfiniteWords('ab')) # random
            12
        """
        return hash(self.alphabet()) ^ hash('infinite words')

    @cached_method
    def factors(self):
        r"""
        Return the set of finite words on the same alphabet.

        EXAMPLES::

            sage: InfiniteWords('ab').factors()
            Finite words over {'a', 'b'}
        """
        return FiniteWords(self.alphabet())

    def shift(self):
        r"""
        Return itself.

        EXAMPLES::

            sage: InfiniteWords('ab').shift()
            Infinite words over {'a', 'b'}
        """
        return self

    @lazy_attribute
    def _element_classes(self):
        r"""
        Return a dictionary that gives the class of the element of ``self``.

        The word may be finite, infinite or of unknown length.
        Its data may be string, list, tuple, a callable or an iterable.
        For callable and iterable, the data may be cached.

        EXAMPLES:

        Once you get the class, it can be used to create a new word::

            sage: W = InfiniteWords([0,1,2])
            sage: cls = W._element_classes['iter_with_caching']
            sage: from itertools import count
            sage: w = cls(W, (i%3 for i in count()))
            sage: type(w)
            <class 'sage.combinat.words.word.InfiniteWord_iter_with_caching'>
            sage: w
            word: 0120120120120120120120120120120120120120...
            sage: w.parent()
            Infinite words over {0, 1, 2}

        TESTS::

            sage: d = InfiniteWords()._element_classes
            sage: type(d)
            <class 'dict'>
            sage: len(d)
            4
            sage: e = InfiniteWords('abcdefg')._element_classes
            sage: d == e
            True
        """
        import sage.combinat.words.word as word
        return {
            'callable_with_caching': word.InfiniteWord_callable_with_caching,
            'callable': word.InfiniteWord_callable,
            'iter_with_caching': word.InfiniteWord_iter_with_caching,
            'iter': word.InfiniteWord_iter}

    def random_element(self, *args, **kwds):
        r"""
        Return a random infinite word.

        EXAMPLES::

            sage: W = InfiniteWords('ab')
            sage: W.random_element() # random
            word: abbbabbaabbbabbabbaabaabbabbbbbbbbaabbbb...

            sage: W = InfiniteWords(ZZ)
            sage: W.random_element(x=2,y=4) # random
            word: 3333223322232233333223323223222233233233...
        """
        rd = self.alphabet().random_element
        from itertools import count
        return self._word_from_iter(rd(*args, **kwds) for i in count())

    def _word_from_word(self, data):
        r"""
        Return a word from a word.

        The data is assumed to be ok, no check is performed.

        INPUT:

        - ``data`` -- word

        EXAMPLES::

            sage: W = InfiniteWords([0,1,2])
            sage: w = W(words.FibonacciWord())
            sage: w
            word: 0100101001001010010100100101001001010010...
            sage: w.parent() is W
            True
            sage: z = W._word_from_word(w)
            sage: w is z
            True
        """
        ####################
        # If `data` is already a word and if its parent is self, then
        # return `data` (no matter what the parameter length, datatype)
        ###########################
        if data.parent() is self or data.parent() == self:
            return data
        elif data.length() != Infinity:
            raise ValueError("cannot build an infinite word from a finite one")

        ###########################
        # Otherwise, if self is not the parent of `data`, then we try to
        # recover the data, the length and the datatype of the input `data`
        ###########################
        from sage.combinat.words.word_infinite_datatypes import (WordDatatype_callable,
                                                                 WordDatatype_iter)
        if isinstance(data, WordDatatype_callable):
            data = data._func
            return self._word_from_callable(data, caching=False)
        elif isinstance(data, WordDatatype_iter):
            data = iter(data)
            return self._word_from_iter(data, caching=False)
        else:
            raise TypeError("any instance of Word_class must be an instance of WordDatatype")

    def _word_from_callable(self, data, caching=True):
        r"""
        Return a word represented by a callable.

        The data is assumed to be ok, no check is performed.

        INPUT:

        - ``data`` -- callable

        - ``caching`` -- boolean (default: ``True``); whether to keep a cache
          of the letters computed by the callable

        EXAMPLES::

            sage: W = InfiniteWords([0,1,2])
            sage: f = lambda n : n % 3
            sage: W._word_from_callable(f)
            word: 0120120120120120120120120120120120120120...
        """
        wc = '_with_caching' if caching else ""
        return self._element_classes['callable' + wc](self, data, Infinity)

    def _word_from_iter(self, data, caching=True):
        r"""
        Return a word represented by an iterator.

        The data is assumed to be ok, no check is performed.

        INPUT:

        - ``data`` -- iterable

        - ``caching`` -- boolean (default: ``True``); whether to keep a cache
          of the letters computed by the iterator

        EXAMPLES::

            sage: W = InfiniteWords([0,1,2])
            sage: from itertools import count
            sage: W._word_from_iter((i % 3 for i in count()))
            word: 0120120120120120120120120120120120120120...
        """
        wc = '_with_caching' if caching else ""
        return self._element_classes['iter' + wc](self, data, Infinity)

    def __call__(self, data=None, datatype=None, caching=True, check=True):
        r"""
        Construct a new word object with parent ``self``.

        INPUT:

        - ``data`` -- iterator or a callable

        - ``datatype`` -- (default: ``None``) ``None``, "iter", "callable" or
          "pickled_function"; if ``None``, then the function tries to guess
          this from the data

        - ``caching`` -- boolean (default: ``True``); whether to keep a
          cache of the letters computed by an iterator or callable

        - ``check`` -- boolean (default: ``True``); whether to check if
          the 40 first letters are in the parent alphabet. This is a
          check done to test for small programming errors. Since we also
          support infinite words, we cannot really implement a more
          accurate check.

        .. NOTE::

           The check makes this method about 10 times slower (20µs instead
           of 2µs), so make sure to set it to False if you know the
           alphabet is OK. Fast creation (about 1µs) of a word can be
           done using the class directly (see :meth:`_element_classes`).

        .. WARNING::

           Be careful when defining words using callables and iterators. It
           appears that islice does not pickle correctly causing various errors
           when reloading. Also, most iterators do not support copying and
           should not support pickling by extension.

        EXAMPLES:

        Word with iterator::

            sage: from itertools import count
            sage: InfiniteWords()(count())
            word: 0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,...

        Word with function (a 'callable')::

            sage: f = lambda n : add(Integer(n).digits(2)) % 2
            sage: InfiniteWords()(f)
            word: 0110100110010110100101100110100110010110...

        The fourty first letters of the word are checked if they are in the
        parent alphabet::

            sage: from itertools import count
            sage: InfiniteWords("ab")(("c" if i == 0 else "a" for i in count()))
            Traceback (most recent call last):
            ...
            ValueError: c not in alphabet

        Creation of a word from a word::

            sage: w = InfiniteWords([0,1,2,3])(words.FibonacciWord())
            sage: w
            word: 0100101001001010010100100101001001010010...
            sage: w.parent()
            Infinite words over {0, 1, 2, 3}
            sage: InfiniteWords([0,1,2,3])(w) is w
            True

        Creation of a word from a pickled function::

            sage: f = lambda n : n % 10
            sage: from sage.misc.fpickle import pickle_function
            sage: s = pickle_function(f)
            sage: Word(s, datatype='pickled_function')
            word: 0123456789012345678901234567890123456789...
        """
        if datatype is not None:
            if datatype == 'callable':
                w = self._word_from_callable(data, caching)
            elif datatype == 'iter':
                w = self._word_from_iter(data, caching)
            elif datatype == 'pickled_function':
                from sage.misc.fpickle import unpickle_function
                data = unpickle_function(data)
                w = self._word_from_callable(data, caching)
            else:
                raise ValueError("unknown datatype (={})".format(datatype))

        elif callable(data):
            w = self._word_from_callable(data, caching)

        elif isinstance(data, Iterable):
            from sage.combinat.words.abstract_word import Word_class
            if isinstance(data, Word_class):
                w = self._word_from_word(data)
            else:
                w = self._word_from_iter(data, caching)

        else:
            raise ValueError("cannot guess a datatype from data (=%s); please specify one" % data)

        if check:
            self._check(w)
        return w

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: Words('ab', finite=False)  # indirect doctest
            Infinite words over {'a', 'b'}
        """
        return "Infinite words over {!r}".format(self.alphabet())

    def _an_element_(self):
        r"""
        Return an element of ``self``.

        EXAMPLES::

            sage: W = Words('ac', finite=False); W
            Infinite words over {'a', 'c'}
            sage: W.an_element()
            word: accacaaccaacaccacaacaccaaccacaaccaacacca...

            sage: W = Words(NN, finite=False); W
            Infinite words over Non negative integer semiring
            sage: W.an_element()
            word: 0110100110010110100101100110100110010110...

            sage: W = Words('z', finite=False); W
            Infinite words over {'z'}
            sage: W.an_element()
            word: zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz...
        """
        some_letters = list(self.alphabet().some_elements())
        if len(some_letters) > 1:
            from sage.combinat.words.word_generators import words
            letters = some_letters[:2]
            return self(words.ThueMorseWord(alphabet=letters))
        else:
            letter = some_letters[0]
            return self(lambda n: letter)


class FiniteOrInfiniteWords(AbstractLanguage):
    def __init__(self, alphabet):
        r"""
        INPUT:

        - ``alphabet`` -- the underlying alphabet

        TESTS::

            sage: loads(dumps(Words())) == Words()
            True
        """
        AbstractLanguage.__init__(self, alphabet)

    def __setstate__(self, state):
        r"""
        TESTS::

            sage: import os, tempfile
            sage: W = Words('ab')
            sage: with tempfile.TemporaryDirectory() as d:
            ....:     filename = os.path.join(d, 'test.sobj')
            ....:     W.save(filename)
            ....:     load(filename)
            Finite and infinite words over {'a', 'b'}
        """
        # add a default to support old pickles from #19619
        self._alphabet = state.get('_alphabet', build_alphabet())

    def cardinality(self):
        r"""
        Return the cardinality of this set of words.

        EXAMPLES::

            sage: Words('abcd').cardinality()
            +Infinity
            sage: Words('a').cardinality()
            +Infinity
            sage: Words('').cardinality()
            1
        """
        return self.finite_words().cardinality()

    @lazy_attribute
    def _element_classes(self):
        r"""
        Return the element classes corresponding to words of unknown length.

        EXAMPLES::

            sage: Words('ab')._element_classes
            {'iter': <class 'sage.combinat.words.word.Word_iter'>,
             'iter_with_caching': <class 'sage.combinat.words.word.Word_iter_with_caching'>}
        """
        import sage.combinat.words.word as word
        return {'iter_with_caching': word.Word_iter_with_caching,
                'iter': word.Word_iter}

    def __hash__(self):
        r"""
        TESTS::

            sage: hash(Words('ab')) # random
            12
        """
        return hash(self.alphabet()) ^ hash('words')

    @cached_method
    def finite_words(self):
        r"""
        Return the set of finite words.

        EXAMPLES::

            sage: Words('ab').finite_words()
            Finite words over {'a', 'b'}
        """
        return FiniteWords(self.alphabet())

    factors = finite_words

    @cached_method
    def infinite_words(self):
        r"""
        Return the set of infinite words.

        EXAMPLES::

            sage: Words('ab').infinite_words()
            Infinite words over {'a', 'b'}
        """
        return InfiniteWords(self.alphabet())

    shift = infinite_words

    def iterate_by_length(self, length):
        r"""
        Return an iterator over the words of given length.

        EXAMPLES::

            sage: [w.string_rep() for w in Words('ab').iterate_by_length(3)]
            ['aaa', 'aab', 'aba', 'abb', 'baa', 'bab', 'bba', 'bbb']
        """
        return self.finite_words().iterate_by_length(length)

    def _word_from_word(self, data):
        r"""
        TESTS::

            sage: W = Words('ab')
            sage: w = FiniteWords('abc')('abba')
            sage: W._word_from_word(w)
            word: abba
            sage: _.parent()
            Finite words over {'a', 'b'}

            sage: w = InfiniteWords('abc')(lambda i: 'a')
            sage: W._word_from_word(w)
            word: aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa...
            sage: _.parent()
            Infinite words over {'a', 'b'}
        """
        P = data.parent()
        if P is self or P is self.finite_words() or P is self.infinite_words() or \
           P == self or P == self.finite_words() or P == self.infinite_words():
            return data
        elif data.is_finite():
            return self.finite_words()._word_from_word(data)
        else:
            return self.infinite_words()._word_from_word(data)

    def _word_from_iter(self, data, caching=True):
        r"""
        TESTS::

            sage: W = Words([0,1,2])
            sage: u = Word(iter("abcabc"*100))
            sage: type(u)
            <class 'sage.combinat.words.word.Word_iter_with_caching'>
            sage: u.length() is None
            True

            sage: u = Word(iter("abcabc"))
            sage: type(u)
            <class 'sage.combinat.words.word.FiniteWord_iter_with_caching'>
            sage: u.length()
            6
        """
        wc = '_with_caching' if caching else ''
        cls = self._element_classes['iter' + wc]
        return cls(self, data, None)

    def __call__(self, data=None, length=None, datatype=None, caching=True, check=True):
        r"""
        Construct a new word object with parent ``self``.

        INPUT:

        - ``data`` -- (default: ``None``) list, string, tuple, iterator, ``None``
          (shorthand for []), or a callable defined on [0,1,...,length]

        - ``length`` -- (default: ``None``) this is dependent on the type of data.
          It is ignored for words defined by lists, strings, tuples,
          etc., because they have a naturally defined length.
          For callables, this defines the domain of definition,
          which is assumed to be [0, 1, 2, ..., length-1].
          For iterators: Infinity if you know the iterator will not
          terminate (default); ``'unknown'`` if you do not know whether the
          iterator terminates; ``'finite'`` if you know that the iterator
          terminates, but do not know the length.

        - ``datatype`` -- (default: ``None``) ``None``, "char", "list", "str",
          "tuple", "iter", "callable" or "pickled_function"; if ``None``, then
          the function tries to guess this from the data.

        - ``caching`` -- boolean (default: ``True``); whether to keep a cache
          of the letters computed by an iterator or callable

        - ``check`` -- boolean (default: ``True``); whether to check if
          the 40 first letters are in the parent alphabet. This is a
          check done to test for small programming errors. Since we also
          support infinite words, we cannot really implement a more
          accurate check.

        .. NOTE::

           The check makes this method about 10 times slower (20µs instead
           of 2µs), so make sure to set it to False if you know the
           alphabet is OK. Fast creation (about 1µs) of a word can be
           done using the class directly (see :meth:`_element_classes`).

        .. WARNING::

           Be careful when defining words using callables and iterators. It
           appears that islice does not pickle correctly causing various errors
           when reloading. Also, most iterators do not support copying and
           should not support pickling by extension.

        EXAMPLES:

        Empty word::

            sage: Words()()
            word:

        Word with string::

            sage: Words()("abbabaab")
            word: abbabaab

        Word with string constructed from other types::

            sage: Words()([0,1,1,0,1,0,0,1], datatype='str')
            word: 01101001
            sage: Words()((0,1,1,0,1,0,0,1), datatype='str')
            word: 01101001

        Word with list::

            sage: Words()([0,1,1,0,1,0,0,1])
            word: 01101001

        Word with list constructed from other types::

            sage: Words()("01101001", datatype='list')
            word: 01101001
            sage: Words()((0,1,1,0,1,0,0,1), datatype='list')
            word: 01101001

        Word with tuple::

            sage: Words()((0,1,1,0,1,0,0,1))
            word: 01101001

        Word with tuple constructed from other types::

            sage: Words()([0,1,1,0,1,0,0,1], datatype='tuple')
            word: 01101001
            sage: Words()("01101001", datatype='str')
            word: 01101001

        Word with iterator::

            sage: from itertools import count
            sage: Words()(count())
            word: 0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,...
            sage: Words()(iter("abbabaab")) # iterators default to infinite words
            word: abbabaab
            sage: Words()(iter("abbabaab"), length='unknown')
            word: abbabaab
            sage: Words()(iter("abbabaab"), length='finite')
            word: abbabaab

        Word with function (a 'callable')::

            sage: f = lambda n : add(Integer(n).digits(2)) % 2
            sage: Words()(f)
            word: 0110100110010110100101100110100110010110...
            sage: Words()(f, length=8)
            word: 01101001

        Word over a string with a parent::

            sage: w = Words('abc')("abbabaab"); w
            word: abbabaab
            sage: w.parent()
            Finite words over {'a', 'b', 'c'}

        The fourty first letters of the word are checked if they are in the
        parent alphabet::

            sage: Words("ab")("abca")
            Traceback (most recent call last):
            ...
            ValueError: c not in alphabet
            sage: Words("ab")("abca", check=False)
            word: abca

        The default parent is the combinatorial class of all words::

            sage: w = Words()("abbabaab"); w
            word: abbabaab
            sage: w.parent()
            Finite words over Set of Python objects of class 'object'

        Creation of a word from a word::

            sage: Words([0,1,2,3])(Words([2,3])([2,2,2,3,3,2]))
            word: 222332
            sage: _.parent()
            Finite words over {0, 1, 2, 3}

        ::

            sage: Words([3,2,1])(Words([2,3])([2,2,2,3,3,2]))
            word: 222332
            sage: _.parent()
            Finite words over {3, 2, 1}

        Construction of a word from a word when the parents are the same::

            sage: W = Words()
            sage: w = W(range(8))
            sage: z = W(w)
            sage: w is z
            True

        Construction of a word path from a finite word::

            sage: W = Words('abcd')
            sage: P = WordPaths('abcd')                                                 # needs sage.modules
            sage: w = W('aaab')
            sage: P(w)                                                                  # needs sage.modules
            Path: aaab

        Construction of a word path from a Christoffel word::

            sage: w = words.ChristoffelWord(5,8)
            sage: w
            word: 0010010100101
            sage: P = WordPaths([0,1,2,3])                                              # needs sage.modules
            sage: P(w)                                                                  # needs sage.modules
            Path: 0010010100101

        Construction of a word represented by a list from a word
        represented by a str ::

            sage: w = Word('ababbbabab')
            sage: type(w)
            <class 'sage.combinat.words.word.FiniteWord_str'>
            sage: z = Word(w, datatype='list')
            sage: type(z)
            <class 'sage.combinat.words.word.FiniteWord_list'>
            sage: y = Word(w, alphabet='abc', datatype='list')
            sage: type(y)
            <class 'sage.combinat.words.word.FiniteWord_list'>

        Creation of a word from a concatenation of words::

            sage: W = Words()
            sage: w = W() * W('a')
            sage: Z = Words('ab')
            sage: Z(w)
            word: a

        Creation of a word path from a FiniteWord_iter::

            sage: w = words.FibonacciWord()
            sage: f = w[:100]
            sage: P = WordPaths([0,1,2,3])                                              # needs sage.modules
            sage: p = P(f); p                                                           # needs sage.modules
            Path: 0100101001001010010100100101001001010010...
            sage: p.length()                                                            # needs sage.modules
            100

        Creation of a word path from a :class:`FiniteWord_callable`::

            sage: g = Word(lambda n: n%2, length=100)
            sage: P = WordPaths([0,1,2,3])                                              # needs sage.modules
            sage: p = P(g); p                                                           # needs sage.modules
            Path: 0101010101010101010101010101010101010101...
            sage: p.length()                                                            # needs sage.modules
            100

        Creation of a word from a pickled function::

            sage: f = lambda n: n % 10
            sage: from sage.misc.fpickle import pickle_function
            sage: s = pickle_function(f)
            sage: Word(s, datatype='pickled_function')
            word: 0123456789012345678901234567890123456789...

        If the alphabet is a subset of [0, 255], then it uses char as datatype::

            sage: type(Word([0,1,1,2,0], alphabet=list(range(256))))
            <class 'sage.combinat.words.word.FiniteWord_char'>

        If the alphabet is a subset of [0, 255], then the letters must
        convert to an unsigned char. Otherwise an error is raised before
        the check is done::

            sage: type(Word([0,1,1,2,0,257], alphabet=list(range(256))))
            Traceback (most recent call last):
            ...
            OverflowError: value too large to convert to unsigned char
            sage: type(Word([0,1,1,2,0,258], alphabet=list(range(257))))
            Traceback (most recent call last):
            ...
            ValueError: 258 not in alphabet
            sage: type(Word([0,1,1,2,0,103], alphabet=list(range(100))))
            Traceback (most recent call last):
            ...
            ValueError: 103 not in alphabet

        Check that the type is rightly guessed for parking functions which are
        callable::

            sage: p = ParkingFunction([2,2,1])
            sage: Word(p).parent()
            Finite words over Set of Python objects of class 'object'
        """
        # try to guess `length` from the `datatype` or `data` if not given
        if length is None or length == 'unknown':
            if data is None:
                length = 'finite'
            elif datatype in ('callable', 'pickled_function'):
                length = 'infinite'
            elif datatype in ('list', 'char', 'str', 'tuple'):
                length = 'finite'
            elif datatype is None:
                try:
                    length = len(data)
                except TypeError:
                    if callable(data):
                        length = 'infinite'

        # now build finite/infinite or unknown length words
        if length == 'finite' or length in ZZ:
            return self.finite_words()(data, datatype=datatype, length=length, caching=caching, check=check)

        elif length == 'infinite' or length == Infinity:
            return self.infinite_words()(data, datatype=datatype, check=check, caching=caching)

        elif length == 'unknown' or length is None:
            from sage.combinat.words.abstract_word import Word_class
            if isinstance(data, Word_class):
                w = self._word_from_word(data)
            elif isinstance(data, Iterable):
                w = self._word_from_iter(data, caching)
            else:
                raise ValueError("cannot guess a datatype from data (={!r}); please specify one".format(data))

            if check:
                w.parent()._check(w)
            return w

        else:
            raise ValueError("invalid argument length (={!r})".format(length))

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: Words('ab', finite=False)._repr_()
            "Infinite words over {'a', 'b'}"
        """
        return "Finite and infinite words over {!r}".format(self.alphabet())


class Words_n(Parent):
    r"""
    The set of words of fixed length on a given alphabet.
    """
    def __init__(self, words, n):
        r"""
        INPUT:

        - ``words`` -- set of finite words

        - ``n`` -- nonnegative integer

        TESTS::

            sage: Words([0,1], length=-42)
            Traceback (most recent call last):
            ...
            ValueError: n = -42 must be nonnegative
        """
        n = ZZ(n)
        if n < 0:
            raise ValueError("n = {} must be nonnegative".format(n))
        self._words = words
        self._n = n

        Parent.__init__(self, category=Sets(), facade=(words,))

    def __setstate__(self, state):
        r"""
        TESTS::

            sage: import os, tempfile
            sage: W = Words('ab', 10)
            sage: with tempfile.TemporaryDirectory() as d:
            ....:     filename = os.path.join(d, 'test.sobj')
            ....:     W.save(filename)
            ....:     load(filename)
            Words of length 10 over {'a', 'b'}
        """
        # add a default to support old pickles from #19619
        self._n = state.get('_n')
        self._words = state.get('_words', FiniteWords())

    def alphabet(self):
        r"""
        Return the underlying alphabet.

        EXAMPLES::

            sage: Words([0,1], 4).alphabet()
            {0, 1}
        """
        return self._words.alphabet()

    def __call__(self, data, *args, **kwds):
        r"""
        INPUT:

        - all arguments are sent directly to the underlying set of finite words.
          See the documentation there for the actual input.

        TESTS::

            sage: Words(5,3)([1,2,3])
            word: 123
            sage: Words(5,3)([1,2,3,1])
            Traceback (most recent call last):
            ...
            ValueError: wrong length
        """
        if 'length' in kwds:
            if kwds['length'] != self._n:
                raise ValueError("wrong length")
        else:
            kwds['length'] = self._n
        w = self._words(data, *args, **kwds)

        if kwds.get('check', True):
            if w.length() != self._n:
                raise ValueError("wrong length")
        return w

    def list(self):
        r"""
        Return a list of all the words contained in ``self``.

        EXAMPLES::

            sage: Words(0,0).list()
            [word: ]
            sage: Words(5,0).list()
            [word: ]
            sage: Words(['a','b','c'],0).list()
            [word: ]
            sage: Words(5,1).list()
            [word: 1, word: 2, word: 3, word: 4, word: 5]
            sage: Words(['a','b','c'],2).list()
            [word: aa, word: ab, word: ac, word: ba, word: bb, word: bc, word: ca, word: cb, word: cc]
        """
        return list(self)

    def _an_element_(self):
        r"""
        Return an element of ``self``.

        EXAMPLES::

            sage: W = Words(2, 3); W
            Words of length 3 over {1, 2}
            sage: W.an_element()
            word: 121

            sage: W = Words("bac", 7); W
            Words of length 7 over {'b', 'a', 'c'}
            sage: W.an_element()
            word: bacbacb

            sage: W = Words("baczxy", 5); W
            Words of length 5 over {'b', 'a', 'c', 'z', 'x', 'y'}
            sage: W.an_element()
            word: baczx
        """
        letters = list(self.alphabet().some_elements())
        r = self._n % len(letters)
        q = (self._n - r) / len(letters)
        return self(letters * int(q) + letters[:r])

    def random_element(self, *args, **kwds):
        r"""
        Return a random word in this set.

        EXAMPLES::

            sage: W = Words('ab', 4)
            sage: W.random_element()  # random
            word: bbab
            sage: W.random_element() in W
            True

            sage: W = Words(ZZ, 5)
            sage: W.random_element()  # random
            word: 1,2,2,-1,12
            sage: W.random_element() in W
            True

        TESTS::

            sage: _ = Words(GF(5),4).random_element()                                   # needs sage.rings.finite_rings

        Check that :issue:`18283` is fixed::

            sage: w = Words('abc', 5).random_element()
            sage: w.length()
            5
        """
        return self._words.random_element(length=self._n, *args, **kwds)

    def _repr_(self):
        """
        EXAMPLES::

            sage: Words(3,5) # indirect doctest
            Words of length 5 over {1, 2, 3}
        """
        from sage.combinat.words.word_options import word_options
        if word_options['old_repr']:
            return "Words over {} of length {}".format(self.alphabet(), self._n)
        return "Words of length {} over {}".format(self._n, self.alphabet())

    def __contains__(self, x):
        """
        EXAMPLES::

            sage: W = Words(3,5)
            sage: W.an_element() in W
            True

            sage: 2 in Words(length=3)
            False
            sage: [1,'a',3] in Words(length=3)
            False
            sage: [1,2] in Words(length=3)
            False
            sage: "abc" in Words(length=3)
            False
            sage: Words("abc")("ababc") in Words(length=3)
            False
            sage: Words([0,1])([1,0,1]) in Words([0,1], length=3)
            True
        """
        return x in self._words and x.length() == self._n

    def cardinality(self):
        r"""
        Return the number of words of length `n` from alphabet.

        EXAMPLES::

            sage: Words(['a','b','c'], 4).cardinality()
            81
            sage: Words(3, 4).cardinality()
            81
            sage: Words(0,0).cardinality()
            1
            sage: Words(5,0).cardinality()
            1
            sage: Words(['a','b','c'],0).cardinality()
            1
            sage: Words(0,1).cardinality()
            0
            sage: Words(5,1).cardinality()
            5
            sage: Words(['a','b','c'],1).cardinality()
            3
            sage: Words(7,13).cardinality()
            96889010407
            sage: Words(['a','b','c','d','e','f','g'],13).cardinality()
            96889010407
        """
        return self.alphabet().cardinality() ** self._n

    __len__ = cardinality

    def __iter__(self):
        r"""
        TESTS::

            sage: [w for w in Words(['a', 'b'], 2)]
            [word: aa, word: ab, word: ba, word: bb]
            sage: [w for w in Words(['b', 'a'], 2)]
            [word: bb, word: ba, word: ab, word: aa]
            sage: [w for w in Words(['a', 'b'], 0)]
            [word: ]
            sage: [w for w in Words([], 3)]
            []
        """
        return self._words.iterate_by_length(self._n)

    def iterate_by_length(self, length):
        r"""
        All words in this class are of the same length, so use iterator
        instead.

        TESTS::

            sage: W = Words(['a', 'b'], 2)
            sage: list(W.iterate_by_length(2))
            [word: aa, word: ab, word: ba, word: bb]
            sage: list(W.iterate_by_length(1))
            []
        """
        if length == self._n:
            return iter(self)
        else:
            return iter([])


###############
# old pickles #
###############


register_unpickle_override("sage.combinat.words.words", "Words_over_OrderedAlphabet", FiniteOrInfiniteWords)
register_unpickle_override("sage.combinat.words.words", "Words_over_Alphabet", FiniteOrInfiniteWords)
register_unpickle_override("sage.combinat.words.words", "FiniteWords_length_k_over_OrderedAlphabet", Words_n)
register_unpickle_override("sage.combinat.words.words", "FiniteWords_over_OrderedAlphabet", FiniteWords)
register_unpickle_override("sage.combinat.words.words", "InfiniteWords_over_OrderedAlphabet", InfiniteWords)



# sage_setup: distribution = sagemath-combinat
r"""
Combinatorial diagrams

A combinatorial diagram is a collection of cells `(i,j)` indexed by pairs of
natural numbers.

For arbitrary diagrams, see :class:`Diagram`. There are also two other specific
types of diagrams implemented here. They are northwest diagrams
(:class:`NorthwestDiagram`) and Rothe diagrams (:func:`RotheDiagram`, a special
kind of northwest diagram).

AUTHORS:

- Trevor K. Karn (2022-08-01): initial version
"""

# ****************************************************************************
#       Copyright (C) 2022 Trevor K. Karn <karnx018 (at) umn.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from itertools import product

from sage.categories.infinite_enumerated_sets import InfiniteEnumeratedSets
from sage.combinat.composition import Composition
from sage.combinat.partition import Partition
from sage.combinat.permutation import Permutations
from sage.combinat.skew_partition import SkewPartition
from sage.combinat.skew_tableau import SkewTableaux
from sage.combinat.tableau import Tableau
from sage.misc.cachefunc import cached_method
from sage.misc.inherit_comparison import InheritComparisonClasscallMetaclass
from sage.misc.lazy_import import lazy_import
from sage.structure.element import Matrix
from sage.structure.list_clone import ClonableArray
from sage.structure.parent import Parent
from sage.structure.unique_representation import UniqueRepresentation

lazy_import('sage.combinat.tiling', 'Polyomino')


class Diagram(ClonableArray, metaclass=InheritComparisonClasscallMetaclass):
    r"""
    Combinatorial diagrams with positions indexed by rows and columns.

    The positions are indexed by rows and columns as in a matrix. For example,
    a Ferrers diagram is a diagram obtained from a partition
    `\lambda = (\lambda_0, \lambda_1, \ldots, \lambda_{\ell})`, where the
    cells are in rows `i` for `0 \leq i \leq \ell` and the cells in row `i`
    consist of `(i,j)` for `0 \leq j < \lambda_i`. In English notation, the
    indices are read from top left to bottom right as in a matrix.

    Indexing conventions are the same as
    :class:`~sage.combinat.partition.Partition`. Printing the diagram of a
    partition, however, will always be in English notation.

    EXAMPLES:

    To create an arbitrary diagram, pass a list of all cells::

        sage: from sage.combinat.diagram import Diagram
        sage: cells = [(0,0), (0,1), (1,0), (1,1), (4,4), (4,5), (4,6), (5,4), (7, 6)]
        sage: D = Diagram(cells); D
        [(0, 0), (0, 1), (1, 0), (1, 1), (4, 4), (4, 5), (4, 6), (5, 4), (7, 6)]

    We can visualize the diagram by printing ``O``'s and ``.``'s. ``O``'s are
    present in the cells which are present in the diagram and a ``.`` represents
    the absence of a cell in the diagram::

        sage: D.pp()
        O O . . . . .
        O O . . . . .
        . . . . . . .
        . . . . . . .
        . . . . O O O
        . . . . O . .
        . . . . . . .
        . . . . . . O

    We can also check if certain cells are contained in a given diagram::

        sage: (1, 0) in D
        True
        sage: (2, 2) in D
        False

    If you know that there are entire empty rows or columns at the end of the
    diagram, you can manually pass them with keyword arguments ``n_rows=`` or
    ``n_cols=``::

        sage: Diagram([(0,0), (0,3), (2,2), (2,4)]).pp()
        O . . O .
        . . . . .
        . . O . O
        sage: Diagram([(0,0), (0,3), (2,2), (2,4)], n_rows=6, n_cols=6).pp()
        O . . O . .
        . . . . . .
        . . O . O .
        . . . . . .
        . . . . . .
        . . . . . .
    """
    @staticmethod
    def __classcall_private__(self, cells, n_rows=None, n_cols=None, check=True):
        r"""
        Normalize the input so that it lives in the correct parent.

        EXAMPLES::

            sage: from sage.combinat.diagram import Diagram
            sage: D = Diagram([(0,0), (0,3), (2,2), (2,4)])
            sage: D.parent()
            Combinatorial diagrams
        """
        return Diagrams()(cells, n_rows, n_cols, check)

    def __init__(self, parent, cells, n_rows=None, n_cols=None, check=True):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: from sage.combinat.diagram import Diagram
            sage: D1 = Diagram([(0,2),(0,3),(1,1),(3,2)])
            sage: D1.cells()
            [(0, 2), (0, 3), (1, 1), (3, 2)]
            sage: D1.nrows()
            4
            sage: D1.ncols()
            4
            sage: TestSuite(D1).run()

        We can specify the number of rows and columns explicitly,
        in case they are supposed to be empty::

            sage: D2 = Diagram([(0,2),(0,3),(1,1),(3,2)], n_cols=5)
            sage: D2.cells()
            [(0, 2), (0, 3), (1, 1), (3, 2)]
            sage: D2.ncols()
            5
            sage: D2.pp()
            . . O O .
            . O . . .
            . . . . .
            . . O . .
            sage: TestSuite(D2).run()
        """
        self._cells = frozenset(cells)

        if self._cells:
            # minimum possible number of rows/cols
            N_rows = max(c[0] for c in self._cells)
            N_cols = max(c[1] for c in self._cells)
        else:  # if there are no cells
            N_rows = -1
            N_cols = -1

        if n_rows is not None:
            if n_rows <= N_rows:
                raise ValueError('n_rows is too small')
            self._n_rows = n_rows
        else:
            self._n_rows = N_rows + 1
        if n_cols is not None:
            if n_cols <= N_cols:
                raise ValueError('n_cols is too small')
            self._n_cols = n_cols
        else:
            self._n_cols = N_cols + 1

        self._n_nonempty_rows = len(set(i for i, j in self._cells))
        self._n_nonempty_cols = len(set(j for i, j in self._cells))

        ClonableArray.__init__(self, parent, sorted(cells), check)

    def pp(self):
        r"""
        Return a visualization of the diagram.

        Cells which are present in the
        diagram are filled with a ``O``. Cells which are not present in the
        diagram are filled with a ``.``.

        EXAMPLES::

            sage: from sage.combinat.diagram import Diagram
            sage: Diagram([(0,0), (0,3), (2,2), (2,4)]).pp()
            O . . O .
            . . . . .
            . . O . O
            sage: Diagram([(0,0), (0,3), (2,2), (2,4)], n_rows=6, n_cols=6).pp()
            O . . O . .
            . . . . . .
            . . O . O .
            . . . . . .
            . . . . . .
            . . . . . .
            sage: Diagram([]).pp()
            -
        """
        if self._n_rows == 0 or self._n_cols == 0:
            print('-')
            return
        print("\n".join(self._pretty_print()))

    def _ascii_art_(self):
        r"""
        Return a visualization of the diagram.

        Cells which are present in the
        diagram are filled with a ``O``. Cells which are not present in the
        diagram are filled with a ``.``.

        EXAMPLES::

            sage: from sage.combinat.diagram import Diagram
            sage: ascii_art(Diagram([(0,0), (0,3), (2,2), (2,4)]))
            O . . O .
            . . . . .
            . . O . O
            sage: ascii_art(Diagram([(0,0), (0,3), (2,2), (2,4)], n_rows=6, n_cols=6))
            O . . O . .
            . . . . . .
            . . O . O .
            . . . . . .
            . . . . . .
            . . . . . .
            sage: ascii_art(Diagram([]))
            -
        """
        from sage.typeset.ascii_art import ascii_art
        if self._n_rows == 0 or self._n_cols == 0:
            return ascii_art("-")
        return ascii_art("\n".join(self._pretty_print()))

    def _unicode_art_(self):
        r"""
        Return a unicode visualization of the diagram.

        Cells which are present in the
        diagram are filled with a crossed box. Cells which are not present in the
        diagram are filled with an empty box.

        EXAMPLES::

            sage: from sage.combinat.diagram import Diagram
            sage: unicode_art(Diagram([(0,0), (0,3), (2,2), (2,4)]))
            ┌─┬─┬─┬─┬─┐
            │X│ │ │X│ │
            ├─┼─┼─┼─┼─┤
            │ │ │ │ │ │
            ├─┼─┼─┼─┼─┤
            │ │ │X│ │X│
            └─┴─┴─┴─┴─┘
            sage: unicode_art(Diagram([(0,0), (0,3), (2,2), (2,4)], n_rows=6, n_cols=6))
            ┌─┬─┬─┬─┬─┬─┐
            │X│ │ │X│ │ │
            ├─┼─┼─┼─┼─┼─┤
            │ │ │ │ │ │ │
            ├─┼─┼─┼─┼─┼─┤
            │ │ │X│ │X│ │
            ├─┼─┼─┼─┼─┼─┤
            │ │ │ │ │ │ │
            ├─┼─┼─┼─┼─┼─┤
            │ │ │ │ │ │ │
            ├─┼─┼─┼─┼─┼─┤
            │ │ │ │ │ │ │
            └─┴─┴─┴─┴─┴─┘
            sage: unicode_art(Diagram([]))
            ∅
        """
        from sage.typeset.unicode_art import unicode_art
        if self._n_rows == 0 or self._n_cols == 0:
            return unicode_art("∅")

        ndivs = self._n_cols - 1
        cell = "│X"
        empty = "│ "
        it = self._pretty_print(cell, empty)
        ret = "┌─" + "┬─"*ndivs + "┐"
        ret += "\n" + next(it) + "│"
        for row in it:
            ret += "\n├─" + "┼─"*ndivs + "┤"
            ret += "\n" + row + "│"
        ret += "\n└─" + "┴─"*ndivs + "┘"
        return unicode_art(ret)

    def _pretty_print(self, cell='O ', empty='. '):
        r"""
        Return a visualization of the diagram.

        Cells which are present in the
        diagram are filled with ``cell``. Cells which are not present in the
        diagram are filled with ``empty``.

        EXAMPLES::

            sage: from sage.combinat.diagram import Diagram
            sage: "\n".join(Diagram([(0,0), (0,3), (2,2), (2,4)])._pretty_print('x ','. '))
            'x . . x . \n. . . . . \n. . x . x '
            sage: "\n".join(Diagram([(0,0), (0,3), (2,2), (2,4)], n_rows=6, n_cols=6)._pretty_print('x ','. '))
            'x . . x . . \n. . . . . . \n. . x . x . \n. . . . . . \n. . . . . . \n. . . . . . '
        """
        for i in range(self._n_rows):
            output_str = ''
            for j in range(self._n_cols):
                if (i, j) in self:
                    output_str += cell
                else:
                    output_str += empty
            yield output_str

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: from sage.combinat.diagram import Diagram
            sage: latex(Diagram([]))
            {\emptyset}
            sage: latex(Diagram([(0,0), (0,3), (2,2), (2,4)]))
            {\def\lr#1{\multicolumn{1}{|@{\hspace{.6ex}}c@{\hspace{.6ex}}|}{\raisebox{-.3ex}{$#1$}}}
            \raisebox{-.6ex}{$\begin{array}[b]{*{5}{p{0.6ex}}}\cline{1-1}\cline{4-4}
            \lr{\phantom{x}}&&&\lr{\phantom{x}}&\\\cline{1-1}\cline{4-4}
            &&&&\\\cline{3-3}\cline{5-5}
            &&\lr{\phantom{x}}&&\lr{\phantom{x}}\\\cline{3-3}\cline{5-5}
            \end{array}$}
            }
        """
        if self._n_rows == 0 or self._n_cols == 0:
            return "{\\emptyset}"

        lr = r'\def\lr#1{\multicolumn{1}{|@{\hspace{.6ex}}c@{\hspace{.6ex}}|}{\raisebox{-.3ex}{$#1$}}}'

        array = [[("\\phantom{x}" if (i, j) in self else None)
                  for j in range(self._n_cols)]
                 for i in range(self._n_rows)]

        def end_line(r):
            # give the line ending to row ``r``
            if r == 0:
                return "".join(r'\cline{%s-%s}' % (i+1, i+1)
                               for i, j in enumerate(array[0]) if j is not None)
            elif r == len(array):
                return r"\\" + "".join(r'\cline{%s-%s}' % (i+1, i+1)
                                       for i, j in enumerate(array[r-1]) if j is not None)
            else:
                out = r"\\" + "".join(r'\cline{%s-%s}' % (i+1, i+1)
                                      for i, j in enumerate(array[r-1]) if j is not None)
                out += "".join(r'\cline{%s-%s}' % (i+1, i+1)
                               for i, j in enumerate(array[r]) if j is not None)
                return out

        tex = r'\raisebox{-.6ex}{$\begin{array}[b]{*{%s}{p{0.6ex}}}' % (max(map(len, array)))
        tex += end_line(0)+'\n'
        for r in range(len(array)):
            tex += '&'.join('' if c is None else r'\lr{%s}' % (c,) for c in array[r])
            tex += end_line(r+1)+'\n'
        return '{%s\n%s\n}' % (lr, tex+r'\end{array}$}')

    def number_of_rows(self):
        r"""
        Return the total number of rows of ``self``.

        EXAMPLES:

        The following example has three rows which are filled, but they
        are contained in rows 0 to 3 (for a total of four)::

            sage: from sage.combinat.diagram import Diagram
            sage: D1 = Diagram([(0,2),(0,3),(1,1),(3,2)])
            sage: D1.number_of_rows()
            4
            sage: D1.nrows()
            4

        The total number of rows includes including those which are empty.
        We can also include empty rows at the end::

            sage: from sage.combinat.diagram import Diagram
            sage: D = Diagram([(0,2),(0,3),(1,1),(3,2)], n_rows=6)
            sage: D.number_of_rows()
            6
            sage: D.pp()
            . . O O
            . O . .
            . . . .
            . . O .
            . . . .
            . . . .
        """
        return self._n_rows

    nrows = number_of_rows

    def number_of_cols(self):
        r"""
        Return the total number of rows of ``self``.

        EXAMPLES:

        The following example has three columns which are filled, but they
        are contained in rows 0 to 3 (for a total of four)::

            sage: from sage.combinat.diagram import Diagram
            sage: D = Diagram([(0,2),(0,3),(1,1),(3,2)])
            sage: D.number_of_cols()
            4
            sage: D.ncols()
            4

        We can also include empty columns at the end::

            sage: from sage.combinat.diagram import Diagram
            sage: D = Diagram([(0,2),(0,3),(1,1),(3,2)], n_cols=6)
            sage: D.number_of_cols()
            6
            sage: D.pp()
            . . O O . .
            . O . . . .
            . . . . . .
            . . O . . .
        """
        return self._n_cols

    ncols = number_of_cols

    def cells(self):
        r"""
        Return a ``list`` of the cells contained in the diagram ``self``.

        EXAMPLES::

            sage: from sage.combinat.diagram import Diagram
            sage: D1 = Diagram([(0,2),(0,3),(1,1),(3,2)])
            sage: D1.cells()
            [(0, 2), (0, 3), (1, 1), (3, 2)]
        """
        return sorted(self._cells)

    def number_of_cells(self):
        r"""
        Return the total number of cells contained in the diagram ``self``.

        EXAMPLES::

            sage: from sage.combinat.diagram import Diagram
            sage: D1 = Diagram([(0,2),(0,3),(1,1),(3,2)])
            sage: D1.number_of_cells()
            4
            sage: D1.n_cells()
            4
        """
        return len(self._cells)

    n_cells = number_of_cells

    size = number_of_cells

    def check(self):
        r"""
        Check that this is a valid diagram.

        EXAMPLES::

            sage: from sage.combinat.diagram import Diagram
            sage: D = Diagram([(0,0), (0,3), (2,2), (2,4)])
            sage: D.check()

        In the next two examples, a bad diagram is passed.
        The first example fails because one cell is indexed by negative
        integers::

            sage: D = Diagram([(0,0), (0,-3), (2,2), (2,4)])
            Traceback (most recent call last):
            ...
            ValueError: diagrams must be indexed by nonnegative integers

        The next example fails because one cell is indexed by rational
        numbers::

            sage: D = Diagram([(0,0), (0,3), (2/3,2), (2,4)])
            Traceback (most recent call last):
            ...
            ValueError: diagrams must be indexed by nonnegative integers
        """
        from sage.sets.non_negative_integers import NonNegativeIntegers
        NN = NonNegativeIntegers()
        if not all(i in NN for c in self._cells for i in c):
            raise ValueError("diagrams must be indexed by nonnegative integers")

    def specht_module(self, base_ring=None):
        r"""
        Return the Specht module corresponding to ``self``.

        EXAMPLES::

            sage: from sage.combinat.diagram import Diagram
            sage: D = Diagram([(0,0), (1,1), (2,2), (2,3)])
            sage: SM = D.specht_module(QQ)                                              # needs sage.modules
            sage: s = SymmetricFunctions(QQ).s()                                        # needs sage.modules
            sage: s(SM.frobenius_image())                                               # needs sage.modules
            s[2, 1, 1] + s[2, 2] + 2*s[3, 1] + s[4]
        """
        from sage.combinat.specht_module import SpechtModule
        from sage.combinat.symmetric_group_algebra import SymmetricGroupAlgebra
        if base_ring is None:
            from sage.rings.rational_field import QQ
            base_ring = QQ
        R = SymmetricGroupAlgebra(base_ring, len(self))
        return SpechtModule(R, self)

    def specht_module_dimension(self, base_ring=None):
        r"""
        Return the dimension of the Specht module corresponding to ``self``.

        INPUT:

        - ``base_ring`` -- (default: `\QQ`) the base ring

        EXAMPLES::

            sage: from sage.combinat.diagram import Diagram
            sage: D = Diagram([(0,0), (1,1), (2,2), (2,3)])
            sage: D.specht_module_dimension()                                           # needs sage.modules
            12
            sage: D.specht_module(QQ).dimension()                                       # needs sage.modules
            12
        """
        from sage.combinat.specht_module import specht_module_rank
        return specht_module_rank(self, base_ring)

    @cached_method
    def essential_set(self):
        r"""
        Return the essential set of ``self`` as defined by Fulton.

        Let `D` be a diagram. Then the *essential set* of `D` are the
        cells `(i, j) \in D` such that `(i+1, j) \notin D` and
        `(i, j+1) \notin D`; that is, the maximally southwest elements
        in each connected component of `D`.

        EXAMPLES::

            sage: w = Permutation([2, 1, 5, 4, 3])
            sage: D = w.rothe_diagram()
            sage: D.essential_set()
            ((0, 0), (2, 3), (3, 2))
        """
        ret = [c for c in self._cells if (c[0]+1, c[1]) not in self._cells
               and (c[0], c[1]+1) not in self._cells]
        ret.sort()
        return tuple(ret)


class Diagrams(UniqueRepresentation, Parent):
    r"""
    The class of combinatorial diagrams.

    A *combinatorial diagram* is a set of cells indexed by pairs of natural
    numbers. Calling an instance of :class:`Diagrams` is one way to construct
    diagrams.

    EXAMPLES::

        sage: from sage.combinat.diagram import Diagrams
        sage: Dgms = Diagrams()
        sage: D = Dgms([(0,0), (0,3), (2,2), (2,4)])
        sage: D.parent()
        Combinatorial diagrams
    """

    def __init__(self, category=None):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: from sage.combinat.diagram import Diagrams
            sage: Dgms = Diagrams(); Dgms
            Combinatorial diagrams

        TESTS::

            sage: TestSuite(Dgms).run()
        """
        Parent.__init__(self, category=InfiniteEnumeratedSets().or_subcategory(category))

    def __iter__(self):
        r"""
        Iterate over ``self``.

        EXAMPLES::

            sage: from sage.combinat.diagram import Diagrams
            sage: I = iter(Diagrams())
            sage: for i in range(10):
            ....:   print(next(I))
            []
            [(0, 0)]
            [(1, 0)]
            [(0, 0), (1, 0)]
            [(0, 1)]
            [(0, 0), (0, 1)]
            [(0, 1), (1, 0)]
            [(0, 0), (0, 1), (1, 0)]
            [(2, 0)]
            [(0, 0), (2, 0)]
            sage: next(I).parent()
            Combinatorial diagrams

            sage: from sage.combinat.diagram import NorthwestDiagrams
            sage: I = iter(NorthwestDiagrams())
            sage: for i in range(20):
            ....:   print(next(I))
            []
            [(0, 0)]
            [(1, 0)]
            [(0, 0), (1, 0)]
            [(0, 1)]
            [(0, 0), (0, 1)]
            [(0, 0), (0, 1), (1, 0)]
            [(2, 0)]
            [(0, 0), (2, 0)]
            [(1, 0), (2, 0)]
            [(0, 0), (1, 0), (2, 0)]
            [(0, 0), (0, 1), (2, 0)]
            [(0, 0), (0, 1), (1, 0), (2, 0)]
            [(1, 1)]
            [(0, 0), (1, 1)]
            [(1, 0), (1, 1)]
            [(0, 0), (1, 0), (1, 1)]
            [(0, 1), (1, 1)]
            [(0, 0), (0, 1), (1, 1)]
            [(0, 0), (0, 1), (1, 0), (1, 1)]
        """
        from sage.sets.non_negative_integers import NonNegativeIntegers
        from sage.categories.cartesian_product import cartesian_product
        from sage.combinat.subset import subsets
        # the product of positive integers automatically implements an
        # an enumeration which allows us to get out of the first column
        N = NonNegativeIntegers()
        NxN = cartesian_product([N, N])
        X = subsets(NxN)
        while True:
            cells = next(X)
            try:
                yield self.element_class(self, tuple((i, j) for i, j in cells))
            except ValueError:
                # if cells causes the .check method of a
                # subclass to fail, just go to the next one
                pass

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: from sage.combinat.diagram import Diagrams
            sage: Dgms = Diagrams(); Dgms
            Combinatorial diagrams
        """
        return 'Combinatorial diagrams'

    def _element_constructor_(self, cells, n_rows=None, n_cols=None, check=True):
        r"""
        Construct an element of ``self``.

        EXAMPLES::

            sage: from sage.combinat.diagram import Diagrams
            sage: Dgms = Diagrams()
            sage: Dgms([(0,1),(2,2)]).pp()
            . O .
            . . .
            . . O


            sage: from sage.combinat.tiling import Polyomino                            # needs sage.modules
            sage: p = Polyomino([(0,0),(1,0),(1,1),(1,2)])                              # needs sage.modules
            sage: Dgms(p).pp()                                                          # needs sage.modules
            O . .
            O O O

            sage: from sage.combinat.composition import Composition
            sage: a = Composition([4,2,0,2,4])
            sage: Dgms(a).pp()
            O O O O
            O O . .
            . . . .
            O O . .
            O O O O

            sage: M = Matrix([[1,1,1,1],[1,1,0,0],[0,0,0,0],[1,1,0,0],[1,1,1,1]])       # needs sage.modules
            sage: Dgms(M).pp()                                                          # needs sage.modules
            O O O O
            O O . .
            . . . .
            O O . .
            O O O O

        TESTS::

            sage: TestSuite(Dgms).run()
        """
        if isinstance(cells, Polyomino):
            return self.from_polyomino(cells)
        if isinstance(cells, Composition):
            return self.from_composition(cells)
        if isinstance(cells, Matrix):
            return self.from_zero_one_matrix(cells)

        return self.element_class(self, cells, n_rows, n_cols, check)

    def _an_element_(self):
        r"""
        Return an element of ``self``.

        EXAMPLES::

            sage: from sage.combinat.diagram import Diagrams
            sage: Dgms = Diagrams()
            sage: D = Dgms.an_element(); D
            [(0, 2), (1, 1), (2, 3)]
            sage: D.pp()
            . . O .
            . O . .
            . . . O
        """
        return self([(0, 2), (1, 1), (2, 3)])

    def from_polyomino(self, p):
        r"""
        Create the diagram corresponding to a 2d
        :class:`~sage.combinat.tiling.Polyomino`

        EXAMPLES::

            sage: from sage.combinat.tiling import Polyomino                            # needs sage.modules
            sage: p = Polyomino([(0,0),(1,0),(1,1),(1,2)])                              # needs sage.modules
            sage: from sage.combinat.diagram import Diagrams
            sage: Diagrams()(p).pp()                                                    # needs sage.modules
            O . .
            O O O

        We can also call this method directly::

            sage: Diagrams().from_polyomino(p).pp()                                     # needs sage.modules
            O . .
            O O O

        This only works for a 2d :class:`~sage.combinat.tiling.Polyomino`::

            sage: p = Polyomino([(0,0,0), (0,1,0), (1,1,0), (1,1,1)], color='blue')     # needs sage.modules
            sage: Diagrams().from_polyomino(p)                                          # needs sage.modules
            Traceback (most recent call last):
            ...
            ValueError: the polyomino must be 2 dimensional
        """
        if not p._dimension == 2:
            raise ValueError("the polyomino must be 2 dimensional")
        cells = list(map(tuple, p))
        return self.element_class(self, cells)

    def from_composition(self, alpha):
        r"""
        Create the diagram corresponding to a weak composition `\alpha \vDash n`.

        EXAMPLES::

            sage: alpha = Composition([3,0,2,1,4,4])
            sage: from sage.combinat.diagram import Diagrams
            sage: Diagrams()(alpha).pp()
            O O O .
            . . . .
            O O . .
            O . . .
            O O O O
            O O O O
            sage: Diagrams().from_composition(alpha).pp()
            O O O .
            . . . .
            O O . .
            O . . .
            O O O O
            O O O O
        """
        cells = []
        for i, n in enumerate(alpha):
            cells.extend((i, j) for j in range(n))
        return self.element_class(self, cells, check=False)

    def from_zero_one_matrix(self, M, check=True):
        r"""
        Get a diagram from a matrix with entries in `\{0, 1\}`, where
        positions of cells are indicated by the `1`'s.

        EXAMPLES::

            sage: M = matrix([[1,0,1,1],[0,1,1,0]])                                     # needs sage.modules
            sage: from sage.combinat.diagram import Diagrams
            sage: Diagrams()(M).pp()                                                    # needs sage.modules
            O . O O
            . O O .
            sage: Diagrams().from_zero_one_matrix(M).pp()                               # needs sage.modules
            O . O O
            . O O .

            sage: M = matrix([[1, 0, 0], [1, 0, 0], [0, 0, 0]])                         # needs sage.modules
            sage: Diagrams()(M).pp()                                                    # needs sage.modules
            O . .
            O . .
            . . .
        """
        # check matrix is zero-one
        n_rows, n_cols = M.dimensions()

        if check:
            zero = M.base_ring().zero()
            one = M.base_ring().one()
            for i in range(n_rows):
                for j in range(n_cols):
                    if not (M[i, j] == zero or M[i, j] == one):
                        raise ValueError("matrix entries must be 0 or 1")
        cells = [(i, j) for i in range(n_rows) for j in range(n_cols) if M[i, j]]

        return self.element_class(self, cells, n_rows, n_cols, check=False)

    Element = Diagram


####################
# Northwest diagrams
####################

class NorthwestDiagram(Diagram, metaclass=InheritComparisonClasscallMetaclass):
    r"""
    Diagrams with the northwest property.

    A diagram is a set of cells indexed by natural numbers. Such a diagram
    has the *northwest property* if the presence of cells `(i1, j1)` and
    `(i2, j2)` implies the presence of the cell
    `(\min(i1, i2), \min(j1, j2))`. Diagrams with the northwest property are
    called *northwest diagrams*.

    For general diagrams see :class:`Diagram`.

    EXAMPLES::

        sage: from sage.combinat.diagram import NorthwestDiagram
        sage: N = NorthwestDiagram([(0,0), (0, 2), (2,0)])

    To visualize them, use the ``.pp()`` method::

        sage: N.pp()
        O . O
        . . .
        O . .
    """
    @staticmethod
    def __classcall_private__(self, cells, n_rows=None, n_cols=None, check=True):
        """
        Normalize input to ensure a correct parent. This method also allows
        one to specify whether or not to check the northwest property for the
        provided cells.

        EXAMPLES::

            sage: from sage.combinat.diagram import NorthwestDiagram, NorthwestDiagrams
            sage: N1 = NorthwestDiagram([(0,1), (0,2)])
            sage: N2 = NorthwestDiagram([(0,1), (0,3)])
            sage: N1.parent() is N2.parent()
            True
            sage: N3 = NorthwestDiagrams()([(0,1), (0,2)])
            sage: N3.parent() is NorthwestDiagrams()
            True
            sage: N1.parent() is NorthwestDiagrams()
            True
        """
        return NorthwestDiagrams()(cells, n_rows, n_cols, check)

    def check(self):
        r"""
        A diagram has the northwest property if the presence of cells
        `(i1, j1)` and `(i2, j2)` implies the presence of the cell
        `(min(i1, i2), min(j1, j2))`. This method checks if the northwest
        property is satisfied for ``self``

        EXAMPLES::

            sage: from sage.combinat.diagram import NorthwestDiagram
            sage: N = NorthwestDiagram([(0,0), (0,3), (3,0)])
            sage: N.check()

        Here is a non-example::

            sage: notN = NorthwestDiagram([(0,1), (1,0)])  #.check() is implicit
            Traceback (most recent call last):
            ...
            ValueError: diagram is not northwest

        TESTS::

            sage: NorthwestDiagram([(0,1/2)])
            Traceback (most recent call last):
            ...
            ValueError: diagrams must be indexed by nonnegative integers
        """
        from itertools import combinations
        Diagram.check(self)
        if not all((min(i1, i2), min(j1, j2)) in self
                   for (i1, j1), (i2, j2) in combinations(self._cells, 2)):
            raise ValueError("diagram is not northwest")

    def peelable_tableaux(self):
        r"""
        Return the set of peelable tableaux whose diagram is ``self``.

        For a fixed northwest diagram `D`, we say that a Young tableau `T` is
        `D`-peelable if:

        1. the row indices of the cells in the first column of `D` are
           the entries in an initial segment in the first column of `T` and
        2. the tableau `Q` obtained by removing those cells from `T` and playing
           jeu de taquin is `(D-C)`-peelable, where `D-C` is the diagram formed
           by forgetting the first column of `D`.

        Reiner and Shimozono [RS1995]_ showed that the number
        `\operatorname{red}(w)` of reduced words of a permutation `w` may be
        computed using the peelable tableaux of the Rothe diagram `D(w)`.
        Explicitly,

        .. MATH::

            \operatorname{red}(w) = \sum_{T} f_{\operatorname{shape} T},

        where the sum runs over the `D(w)`-peelable tableaux `T` and `f_\lambda`
        is the number of standard Young tableaux of shape `\lambda` (which may
        be computed using the hook-length formula).

        EXAMPLES:

        We can compute the `D`-peelable diagrams for a northwest diagram `D`::

            sage: from sage.combinat.diagram import NorthwestDiagram
            sage: cells = [(0,0), (0,1), (0,2), (1,0), (2,0), (2,2), (2,4),
            ....:          (4,0), (4,2)]
            sage: D = NorthwestDiagram(cells); D.pp()
            O O O . .
            O . . . .
            O . O . O
            . . . . .
            O . O . .
            sage: D.peelable_tableaux()
            {[[1, 1, 1], [2, 3, 3], [3, 5], [5]],
            [[1, 1, 1, 3], [2, 3], [3, 5], [5]]}

        EXAMPLES:

        If the diagram is only one column, there is only one peelable tableau::

            sage: from sage.combinat.diagram import NorthwestDiagram
            sage: NWD = NorthwestDiagram([(0,0), (2,0)])
            sage: NWD.peelable_tableaux()
            {[[1], [3]]}

        From [RS1995]_, we know that there is only one peelable tableau for the
        Rothe diagram of the permutation (in one line notation) `251643`::

            sage: D = NorthwestDiagram([(1, 2), (1, 3), (3, 2), (3, 3), (4, 2)])
            sage: D.pp()
            . . . .
            . . O O
            . . . .
            . . O O
            . . O .

            sage: D.peelable_tableaux()
            {[[2, 2], [4, 4], [5]]}

        Here are all the intermediate steps to compute the peelables for the
        Rothe diagram of (in one-line notation) `64817235`. They are listed from
        deepest in the recursion to the final step. The recursion has depth five
        in this case so we will label the intermediate tableaux by `D_i` where
        `i` is the step in the recursion at which they appear.

        Start with the one that has a single column::

            sage: D5 = NorthwestDiagram([(2,0)]); D5.pp()
            .
            .
            O
            sage: D5.peelable_tableaux()
            {[[3]]}

        Now we know all of the `D_5` peelables, so we can compute the `D_4`
        peelables::

            sage: D4 = NorthwestDiagram([(0, 0), (2,0), (4, 0), (2, 2)])
            sage: D4.pp()
            O . .
            . . .
            O . O
            . . .
            O . .

            sage: D4.peelable_tableaux()
            {[[1, 3], [3], [5]]}

        There is only one `D_4` peelable, so we can compute the `D_3`
        peelables::

            sage: D3 = NorthwestDiagram([(0,0), (0,1), (2, 1), (2, 3), (4,1)])
            sage: D3.pp()
            O O . .
            . . . .
            . O . O
            . . . .
            . O . .

            sage: D3.peelable_tableaux()
            {[[1, 1], [3, 3], [5]], [[1, 1, 3], [3], [5]]}

        Now compute the `D_2` peelables::

            sage: cells = [(0,0), (0,1), (0,2), (1,0), (2,0), (2,2), (2,4),
            ....:          (4,0), (4,2)]
            sage: D2 = NorthwestDiagram(cells); D2.pp()
            O O O . .
            O . . . .
            O . O . O
            . . . . .
            O . O . .

            sage: D2.peelable_tableaux()
            {[[1, 1, 1], [2, 3, 3], [3, 5], [5]],
            [[1, 1, 1, 3], [2, 3], [3, 5], [5]]}

        And the `D_1` peelables::

            sage: cells = [(0,0), (0,1), (0,2), (0,3), (1,0), (1,1), (2,0),
            ....:          (2,1), (2,3), (2,5), (4,0), (4,1), (4,3)]
            sage: D1 = NorthwestDiagram(cells); D1.pp()
            O O O O . .
            O O . . . .
            O O . O . O
            . . . . . .
            O O . O . .

            sage: D1.peelable_tableaux()
            {[[1, 1, 1, 1], [2, 2, 3, 3], [3, 3, 5], [5, 5]],
            [[1, 1, 1, 1, 3], [2, 2, 3], [3, 3, 5], [5, 5]]}

        Which we can use to get the `D` peelables::

            sage: cells = [(0,0), (0,1), (0,2), (0,3), (0,4),
            ....:          (1,0), (1,1), (1,2),
            ....:          (2,0), (2,1), (2,2), (2,4), (2,6),
            ....:                 (4,1), (4,2), (4,4)]
            sage: D = NorthwestDiagram(cells); D.pp()
            O O O O O . .
            O O O . . . .
            O O O . O . O
            . . . . . . .
            . O O . O . .
            sage: D.peelable_tableaux()
            {[[1, 1, 1, 1, 1], [2, 2, 2, 3, 3], [3, 3, 3], [5, 5, 5]],
             [[1, 1, 1, 1, 1], [2, 2, 2, 3, 3], [3, 3, 3, 5], [5, 5]],
             [[1, 1, 1, 1, 1, 3], [2, 2, 2, 3], [3, 3, 3], [5, 5, 5]],
             [[1, 1, 1, 1, 1, 3], [2, 2, 2, 3], [3, 3, 3, 5], [5, 5]]}

        ALGORITHM:

        This implementation uses the algorithm suggested in Remark 25
        of [RS1995]_.

        TESTS:

        Corner case::

            sage: from sage.combinat.diagram import NorthwestDiagram
            sage: D = NorthwestDiagram([])
            sage: D.peelable_tableaux()
            {[]}
        """
        # TODO: There is a condition on the first column (if the rows in Dhat
        # are a subset of the rows in the first column) which simplifies the
        # description without performing JDT, so we should implement that

        # empty diagram case
        if not self:
            return set([Tableau([])])

        # if there is a single column in the diagram then there is only
        # one posslbe peelable tableau.
        if self._n_nonempty_cols == 1:
            return set([Tableau([[i+1] for i, j in self.cells()])])

        first_col = min(j for i, j in self._cells)

        dhat_cells = []
        new_vals_cells = []
        for i, j in self._cells:
            if j != first_col:
                dhat_cells.append((i, j))
            else:
                new_vals_cells.append(i + 1)

        new_vals = sorted(new_vals_cells)

        Dhat = NorthwestDiagram(dhat_cells)
        k = self.n_cells() - Dhat.n_cells()

        peelables = set()

        for Q in Dhat.peelable_tableaux():
            # get the vertical strips
            mu = Q.shape()
            vertical_strip_cells = mu.vertical_border_strip_cells(k)
            for s in vertical_strip_cells:
                sQ = SkewTableaux()(Q)  # sQ is skew - get it?
                # perform the jeu de taquin slides
                for c in s:
                    sQ = sQ.backward_slide(c)
                # create the new tableau by filling the columns
                sQ_new = sQ.to_list()
                for n in range(k):
                    sQ_new[n][0] = new_vals[n]

                T = Tableau(sQ_new)
                if T.is_column_strict():
                    peelables.add(T)

        return peelables


class NorthwestDiagrams(Diagrams):
    r"""
    Diagrams satisfying the northwest property.

    A diagram `D` is a *northwest diagram* if for every two cells `(i_1, j_1)`
    and `(i_2, j_2)` in `D` then there exists the cell
    `(\min(i_1, i_2), \min(j_1, j_2)) \in D`.

    EXAMPLES::

        sage: from sage.combinat.diagram import NorthwestDiagram
        sage: N = NorthwestDiagram([(0,0), (0, 10), (5,0)]); N.pp()
        O . . . . . . . . . O
        . . . . . . . . . . .
        . . . . . . . . . . .
        . . . . . . . . . . .
        . . . . . . . . . . .
        O . . . . . . . . . .

    Note that checking whether or not the northwest property is satisfied is
    automatically checked. The diagram found by adding the cell `(1,1)` to the
    diagram above is *not* a northwest diagram. The cell `(1,0)` should be
    present due to the presence of `(5,0)` and `(1,1)`::

        sage: from sage.combinat.diagram import Diagram
        sage: Diagram([(0, 0), (0, 10), (5, 0), (1, 1)]).pp()
        O . . . . . . . . . O
        . O . . . . . . . . .
        . . . . . . . . . . .
        . . . . . . . . . . .
        . . . . . . . . . . .
        O . . . . . . . . . .
        sage: NorthwestDiagram([(0, 0), (0, 10), (5, 0), (1, 1)])
        Traceback (most recent call last):
        ...
        ValueError: diagram is not northwest

    However, this behavior can be turned off if you are confident that
    you are providing a northwest diagram::

        sage: N = NorthwestDiagram([(0, 0), (0, 10), (5, 0),
        ....:                      (1, 1), (0, 1), (1, 0)],
        ....:                      check=False)
        sage: N.pp()
        O O . . . . . . . . O
        O O . . . . . . . . .
        . . . . . . . . . . .
        . . . . . . . . . . .
        . . . . . . . . . . .
        O . . . . . . . . . .

    Note that arbitrary diagrams which happen to be northwest diagrams
    only live in the parent of :class:`Diagrams`::

        sage: D = Diagram([(0, 0), (0, 10), (5, 0), (1, 1), (0, 1), (1, 0)])
        sage: D.pp()
        O O . . . . . . . . O
        O O . . . . . . . . .
        . . . . . . . . . . .
        . . . . . . . . . . .
        . . . . . . . . . . .
        O . . . . . . . . . .
        sage: from sage.combinat.diagram import NorthwestDiagrams
        sage: D in NorthwestDiagrams()
        False

    Here are some more examples::

        sage: from sage.combinat.diagram import NorthwestDiagram, NorthwestDiagrams
        sage: D = NorthwestDiagram([(0,1), (0,2), (1,1)]); D.pp()
        . O O
        . O .
        sage: NWDgms = NorthwestDiagrams()
        sage: D = NWDgms([(1,1), (1,2), (2,1)]); D.pp()
        . . .
        . O O
        . O .
        sage: D.parent()
        Combinatorial northwest diagrams

    Additionally, there are natural constructions of a northwest diagram
    given the data of a permutation (Rothe diagrams are the prototypical example
    of northwest diagrams), or the data of a partition of an integer, or a
    skew partition.

    The Rothe diagram `D(\omega)` of a permutation `\omega` is specified by
    the cells

    .. MATH::

        D(\omega) = \{(\omega_j, i) : i<j,\, \omega_i > \omega_j \}.

    We can construct one by calling :meth:`rothe_diagram` method on the set
    of all :class:`~sage.combinat.diagram.NorthwestDiagrams`::

        sage: w = Permutations(4)([4,3,2,1])
        sage: NorthwestDiagrams().rothe_diagram(w).pp()
        O O O .
        O O . .
        O . . .
        . . . .

    To turn a Ferrers diagram into a northwest diagram, we may call
    :meth:`from_partition`. This will return a Ferrer's diagram in the
    set of all northwest diagrams. For many use-cases it is probably better
    to get Ferrer's diagrams by the corresponding method on partitions, namely
    :meth:`sage.combinat.partitions.Partitions.ferrers_diagram`::

        sage: mu = Partition([7,3,1,1])
        sage: mu.pp()
        *******
        ***
        *
        *
        sage: NorthwestDiagrams().from_partition(mu).pp()
        O O O O O O O
        O O O . . . .
        O . . . . . .
        O . . . . . .

    It is also possible to turn a Ferrers diagram of a skew partition into a
    northwest diagram, although it is more subtle than just using the skew
    diagram itself. One must first reflect the partition about a vertical axis
    so that the skew partition looks "backwards"::

        sage: mu, nu = Partition([5,4,3,2,1]), Partition([3,2,1])
        sage: s = mu/nu; s.pp()
           **
          **
         **
        **
        *
        sage: NorthwestDiagrams().from_skew_partition(s).pp()
        O O . . .
        . O O . .
        . . O O .
        . . . O O
        . . . . O
    """

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: from sage.combinat.diagram import NorthwestDiagrams
            sage: NWDgms = NorthwestDiagrams(); NWDgms
            Combinatorial northwest diagrams
        """
        return 'Combinatorial northwest diagrams'

    def _an_element_(self):
        r"""
        Return an element of ``self``.

        EXAMPLES::

            sage: from sage.combinat.diagram import NorthwestDiagrams
            sage: NWDgms = NorthwestDiagrams()
            sage: NWD = NWDgms.an_element(); NWD
            [(0, 1), (0, 2), (1, 1), (2, 3)]
            sage: NWD.pp()
            . O O .
            . O . .
            . . . O
            sage: NWD.parent() is NWDgms
            True
        """
        return self([(0, 1), (0, 2), (1, 1), (2, 3)])

    def rothe_diagram(self, w):
        r"""
        Return the Rothe diagram of ``w``.

        We construct a northwest diagram from a permutation by
        constructing its Rothe diagram. Formally, if `\omega` is
        a :class:`~sage.combinat.permutation.Permutation`
        then the Rothe diagram `D(\omega)` is the diagram whose cells are

        .. MATH::

            D(\omega) = \{(\omega_j, i) : i<j,\, \omega_i > \omega_j \}.

        Informally, one can construct the Rothe diagram by starting with all
        `n^2` possible cells, and then deleting the cells `(i, \omega(i))` as
        well as all cells to the right and below. (These are sometimes called
        "death rays".)

        .. SEEALSO::

            :func:`~sage.combinat.diagram.RotheDiagram`

        EXAMPLES::

            sage: from sage.combinat.diagram import NorthwestDiagrams
            sage: w = Permutations(3)([2,1,3])
            sage: NorthwestDiagrams().rothe_diagram(w).pp()
            O . .
            . . .
            . . .
            sage: NorthwestDiagrams().from_permutation(w).pp()
            O . .
            . . .
            . . .

            sage: w = Permutations(8)([2,5,4,1,3,6,7,8])
            sage: NorthwestDiagrams().rothe_diagram(w).pp()
            O . . . . . . .
            O . O O . . . .
            O . O . . . . .
            . . . . . . . .
            . . . . . . . .
            . . . . . . . .
            . . . . . . . .
            . . . . . . . .
        """
        return RotheDiagram(w)

    from_permutation = rothe_diagram

    def from_partition(self, mu):
        r"""
        Return the Ferrer's diagram of ``mu`` as a northwest diagram.

        EXAMPLES::

            sage: mu = Partition([5,2,1]); mu.pp()
            *****
            **
            *
            sage: mu.parent()
            Partitions
            sage: from sage.combinat.diagram import NorthwestDiagrams
            sage: D = NorthwestDiagrams().from_partition(mu)
            sage: D.pp()
            O O O O O
            O O . . .
            O . . . .
            sage: D.parent()
            Combinatorial northwest diagrams

        This will print in English notation even if the notation is set to
        French for the partition::

            sage: Partitions.options.convention="french"
            sage: mu.pp()
            *
            **
            *****
            sage: D.pp()
            O O O O O
            O O . . .
            O . . . .

        TESTS::

            sage: from sage.combinat.diagram import NorthwestDiagrams
            sage: mu = [5, 2, 1]
            sage: D = NorthwestDiagrams().from_partition(mu)
            Traceback (most recent call last):
            ...
            ValueError: mu must be a Partition
        """
        if not isinstance(mu, Partition):
            raise ValueError("mu must be a Partition")
        return self.element_class(self, mu.cells(), check=False)

    def from_skew_partition(self, s):
        r"""
        Get the northwest diagram found by reflecting a skew shape across
        a vertical plane.

        EXAMPLES::

            sage: mu, nu = Partition([3,2,1]), Partition([2,1])
            sage: s = mu/nu; s.pp()
              *
             *
            *
            sage: from sage.combinat.diagram import NorthwestDiagrams
            sage: D = NorthwestDiagrams().from_skew_partition(s)
            sage: D.pp()
            O . .
            . O .
            . . O

            sage: mu, nu = Partition([3,3,2]), Partition([2,2,2])
            sage: s = mu/nu; s.pp()
              *
              *
            sage: NorthwestDiagrams().from_skew_partition(s).pp()
            O . .
            O . .
            . . .

        TESTS::

            sage: mu = Partition([3,2,1])
            sage: NorthwestDiagrams().from_skew_partition(mu)
            Traceback (most recent call last):
            ...
            ValueError: mu must be a SkewPartition
        """
        if not isinstance(s, SkewPartition):
            raise ValueError("mu must be a SkewPartition")

        n_cols = s.outer()[0]
        n_rows = len(s.outer())

        cells = [(i, n_cols - 1 - j) for i, j in s.cells()]

        return self.element_class(self, cells, n_rows, n_cols, check=False)

    def from_parallelogram_polyomino(self, p):
        r"""
        Create the diagram corresponding to a
        :class:`~sage.combinat.parallelogram_polyomino.ParallelogramPolyomino`.

        EXAMPLES::

            sage: p = ParallelogramPolyomino([[0, 0, 1, 0, 0, 0, 1, 1],                 # needs sage.modules
            ....:                              [1, 1, 0, 1, 0, 0, 0, 0]])
            sage: from sage.combinat.diagram import NorthwestDiagrams
            sage: NorthwestDiagrams().from_parallelogram_polyomino(p).pp()              # needs sage.modules
            O O .
            O O O
            . O O
            . O O
            . O O
        """
        from sage.matrix.constructor import Matrix
        M = Matrix(p.get_array())
        return self.from_zero_one_matrix(M)

    Element = NorthwestDiagram


def RotheDiagram(w):
    r"""
    The Rothe diagram of a permutation ``w``.

    EXAMPLES::

        sage: w = Permutations(9)([1, 7, 4, 5, 9, 3, 2, 8, 6])
        sage: from sage.combinat.diagram import RotheDiagram
        sage: D = RotheDiagram(w); D.pp()
        . . . . . . . . .
        . O O O O O . . .
        . O O . . . . . .
        . O O . . . . . .
        . O O . . O . O .
        . O . . . . . . .
        . . . . . . . . .
        . . . . . O . . .
        . . . . . . . . .

    The Rothe diagram is a northwest diagram::

        sage: D.parent()
        Combinatorial northwest diagrams

    Some other examples::

        sage: RotheDiagram([2, 1, 4, 3]).pp()
        O . . .
        . . . .
        . . O .
        . . . .

        sage: RotheDiagram([4, 1, 3, 2]).pp()
        O O O .
        . . . .
        . O . .
        . . . .

    Currently, only elements of the set of
    :class:`sage.combinat.permutations.Permutations` are supported. In
    particular, elements of permutation groups are not supported::

        sage: w = SymmetricGroup(9).an_element()                                        # needs sage.groups
        sage: RotheDiagram(w)                                                           # needs sage.groups
        Traceback (most recent call last):
        ...
        ValueError: w must be a permutation

    TESTS::

        sage: w = Permutations(5)([1,2,3,4,5])
        sage: from sage.combinat.diagram import RotheDiagram
        sage: RotheDiagram(w).pp()
        . . . . .
        . . . . .
        . . . . .
        . . . . .
        . . . . .
    """
    P = Permutations()
    if w not in P:
        raise ValueError('w must be a permutation')
    w = P(w)

    N = w.size()
    winv = w.inverse()
    cells = [c for c in product(range(N), range(N))
             if c[0] + 1 < winv(c[1] + 1) and c[1] + 1 < w(c[0] + 1)]

    return NorthwestDiagram(cells, n_rows=N, n_cols=N, check=False)



# sage_setup: distribution = sagemath-combinat
"""
Abelian Monoid Elements

AUTHORS:

- David Kohel (2005-09)

EXAMPLES:

Recall the example from abelian monoids::

    sage: F = FreeAbelianMonoid(5,names = list("abcde"))
    sage: (a,b,c,d,e) = F.gens()
    sage: a*b^2*e*d
    a*b^2*d*e
    sage: x = b^2*e*d*a^7
    sage: x
    a^7*b^2*d*e
    sage: x.list()
    [7, 2, 0, 1, 1]

The list is a copy, so changing the list does not change the element::

    sage: x.list()[0] = 0
    sage: x
    a^7*b^2*d*e
"""

# ***************************************************************************
#  Copyright (C) 2006 William Stein <wstein@gmail.com>
#  Copyright (C) 2005 David Kohel <kohel@maths.usyd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ***************************************************************************

from cysignals.memory cimport check_allocarray, sig_free
from cysignals.signals cimport sig_on, sig_off
from sage.structure.richcmp cimport rich_to_bool
from sage.rings.integer cimport Integer, _Integer_from_mpz
from sage.libs.gmp.mpz cimport *


def is_FreeAbelianMonoidElement(x):
    r"""
    Queries whether ``x`` is an object of type ``FreeAbelianMonoidElement``.

    INPUT:

    - ``x`` -- an object

    OUTPUT:

    - ``True`` if ``x`` is an object of type ``FreeAbelianMonoidElement``;
      ``False`` otherwise.
    """
    from sage.misc.superseded import deprecation_cython
    deprecation_cython(38184,
                       "The function is_FreeAbelianMonoidElement is deprecated; "
                       "use 'isinstance(..., FreeAbelianMonoidElement)' instead.")
    return isinstance(x, FreeAbelianMonoidElement)


cdef class FreeAbelianMonoidElement(MonoidElement):
    cdef int _init(self, Py_ssize_t n, Parent parent) except -1:
        """
        Initialize the C data structures in this vector. After calling
        this, ``self`` is the identity element represented as a zero vector
        of degree ``n`` with parent ``parent``.

        Only if you call ``__new__`` without a ``parent`` argument, it
        is needed to call this function manually. The only reason to do
        that is for efficiency: calling ``__new__`` without any
        additional arguments besides the type and then calling ``_init``
        is (slightly) more efficient than calling ``__new__`` with a
        ``parent`` argument.
        """
        # Assign variables only when the array is fully initialized
        cdef mpz_t* entries = <mpz_t*>check_allocarray(n, sizeof(mpz_t))
        cdef Py_ssize_t i
        sig_on()
        for i in range(n):
            mpz_init(entries[i])
        sig_off()
        self._element_vector = entries
        self._n = n
        self._parent = parent

    def __cinit__(self, parent=None, x=None):
        """
        C level initialization of ``self``.

        EXAMPLES::

            sage: F = FreeAbelianMonoid(5, 'abcde')
            sage: F(1)
            1
        """
        self._element_vector = NULL
        if parent is None:
            self._n = 0
            return
        self._init(parent.ngens(), <Parent?>parent)

    def __init__(self, parent, x):
        r"""
        Create the element ``x`` of the FreeAbelianMonoid ``parent``.

        EXAMPLES::

            sage: F = FreeAbelianMonoid(5, 'abcde')
            sage: F
            Free abelian monoid on 5 generators (a, b, c, d, e)
            sage: F(1)
            1
            sage: F(2)
            Traceback (most recent call last):
            ...
            TypeError: argument x (= 2) is of the wrong type
            sage: F(int(1))
            1
            sage: a, b, c, d, e = F.gens()
            sage: a^2 * b^3 * a^2 * b^4
            a^4*b^7
            sage: F = FreeAbelianMonoid(5, 'abcde')
            sage: a, b, c, d, e = F.gens()
            sage: a in F
            True
            sage: a*b in F
            True
        """
        cdef Py_ssize_t i
        cdef Integer z

        if isinstance(x, int):
            if (<int> x) == 1:
                return  # the _element_vector is already set to the 0 vector
        elif isinstance(x, Integer):
            if mpz_cmp_si((<Integer> x).value, 1) == 0:
                return  # the _element_vector is already set to the 0 vector
        elif isinstance(x, (list, tuple)):
            if len(x) != self._n:
                raise IndexError("argument length (= %s) must be %s" % (len(x), self._n))
            for i in range(self._n):
                z = Integer(x[i])
                mpz_set(self._element_vector[i], z.value)
            return
        raise TypeError("argument x (= %s) is of the wrong type" % x)

    def __dealloc__(self):
        """
        Deallocate ``self``.

        EXAMPLES::

            sage: F = FreeAbelianMonoid(5, 'abcde')
            sage: a, b, c, d, e = F.gens()
            sage: x = a^2 * b^3
            sage: del x
        """
        cdef Py_ssize_t i
        if self._element_vector:
            for i in range(self._n):
                mpz_clear(self._element_vector[i])
            sig_free(self._element_vector)

    def __copy__(self):
        """
        Return a copy of ``self``.

        EXAMPLES::

            sage: F = FreeAbelianMonoid(5, 'abcde')
            sage: a, b, c, d, e = F.gens()
            sage: x = a^2 * b^3
            sage: copy(x) == x
            True
            sage: copy(x) is x
            False
        """
        cdef FreeAbelianMonoidElement y
        y = self._new_c()
        cdef Py_ssize_t i
        for i in range(self._n):
            mpz_set(y._element_vector[i], self._element_vector[i])
        return y

    def __reduce__(self):
        """
        Used in pickling.

        EXAMPLES::

            sage: F = FreeAbelianMonoid(5, 'abcde')
            sage: a, b, c, d, e = F.gens()
            sage: x = a^2 * b^3
            sage: loads(dumps(x)) == x
            True
        """
        return (self._parent, (self.list(),))

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: F = FreeAbelianMonoid(5, 'abcde')
            sage: F(1)
            1
            sage: a, b, c, d, e = F.gens()
            sage: a^2 * b^3 * a^2 * b^4
            a^4*b^7
        """
        cdef str s = ""
        A = self._parent
        cdef tuple x = A.variable_names()
        cdef mpz_t *v = self._element_vector
        cdef Integer val
        for i in range(self._n):
            val = _Integer_from_mpz(v[i])
            if val == 0:
                continue
            elif val == 1:
                if s:
                    s += "*"
                s += "%s" % x[i]
            else:
                if s:
                    s += "*"
                s += "%s^%s" % (x[i], val)
        if not s:
            s = "1"
        return s

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: F = FreeAbelianMonoid(2, 'alpha,beta')
            sage: latex(F(1))
            1
            sage: a, b = F.gens()
            sage: latex(a^2 * b^3 * a^2 * b^4)
            \alpha^{4} \beta^{7}
        """
        cdef str s = ""
        A = self._parent
        cdef list x = A.latex_variable_names()
        cdef mpz_t *v = self._element_vector
        cdef Integer val
        for i in range(self._n):
            val = _Integer_from_mpz(v[i])
            if val == 0:
                continue
            elif val == 1:
                if s:
                    s += " "
                s += "%s" % x[i]
            else:
                if s:
                    s += " "
                s += "%s^{%s}" % (x[i], val)
        if not s:
            s = "1"
        return s

    cpdef _richcmp_(left, right, int op):
        """
        Rich comparison.

        EXAMPLES::

            sage: F = FreeAbelianMonoid(5, 'abcde')
            sage: F(1)
            1
            sage: a, b, c, d, e = F.gens()
            sage: x = a^2 * b^3
            sage: F(1) < x
            True
            sage: x > b
            True
            sage: x <= a^4
            True
            sage: x != a*b
            True
            sage: a*b == b*a
            True
            sage: x > a^3*b^2
            False
        """
        cdef Py_ssize_t i
        cdef int c
        for i in range(left._n):
            c = mpz_cmp(left._element_vector[i], (<FreeAbelianMonoidElement>right)._element_vector[i])
            if c < 0:
                return rich_to_bool(op, -1)
            elif c > 0:
                return rich_to_bool(op, 1)
        return rich_to_bool(op, 0)

    def __mul__(self, y):
        """
        Multiply ``self`` with ``y``.

        EXAMPLES::

            sage: F = FreeAbelianMonoid(5, 'abcde')
            sage: a, b, c, d, e = F.gens()
            sage: b * a^2 * b^3
            a^2*b^4
        """
        if not isinstance(y, FreeAbelianMonoidElement):
            raise TypeError("argument y (= %s) is of wrong type" % y)
        cdef FreeAbelianMonoidElement s, z, r
        s = self
        r = y
        z = s._new_c()
        cdef Py_ssize_t i
        for i in range(s._n):
            mpz_add(z._element_vector[i], s._element_vector[i], r._element_vector[i])
        return z

    def __pow__(self, n, modulus):
        """
        Raise ``self`` to the power of ``n``.

        AUTHORS:

        - Tom Boothby (2007-08): Replaced O(log n) time, O(n) space
          algorithm with O(1) time and space "algorithm".

        EXAMPLES::

            sage: F = FreeAbelianMonoid(5,names = list("abcde"))
            sage: (a,b,c,d,e) = F.gens()
            sage: x = a*b^2*e*d; x
            a*b^2*d*e
            sage: x^3
            a^3*b^6*d^3*e^3
            sage: x^0
            1
        """
        if modulus is not None:
            raise NotImplementedError("modulus for exponents not implemented")
        cdef Integer val
        if isinstance(n, int):
            val = Integer(n)
        else:
            if not isinstance(n, Integer):
                raise TypeError("argument n (= %s) must be an integer" % (n,))
            val = <Integer> n
        if val < 0:
            raise IndexError("argument n (= %s) must be positive" % val)
        elif val == 1:
            return self
        cdef FreeAbelianMonoidElement s, z
        s = self
        z = s._new_c()
        cdef Py_ssize_t i
        for i in range(s._n):
            mpz_mul(z._element_vector[i], s._element_vector[i], val.value)
        return z

    def __hash__(self):
        """
        Return the hash of ``self``.

        EXAMPLES::

            sage: F = FreeAbelianMonoid(5,names = list("abcde"))
            sage: (a,b,c,d,e) = F.gens()
            sage: x = a*b^2*e*d
            sage: hash(x) == hash(x)
            True
        """
        return hash(tuple(self.list()))

    def list(self):
        r"""
        Return the underlying list used to represent ``self``.

        If this is a monoid in an abelian monoid on `n` generators,
        then this is a list of nonnegative integers of length `n`.

        EXAMPLES::

            sage: F = FreeAbelianMonoid(5, 'abcde')
            sage: (a, b, c, d, e) = F.gens()
            sage: a.list()
            [1, 0, 0, 0, 0]
        """
        cdef Py_ssize_t i
        return [_Integer_from_mpz(self._element_vector[i]) for i in range(self._n)]
